local L = LibStub("AceLocale-3.0"):GetLocale("DrDamage", true)

DrDamage.PlayerAura = {}
DrDamage.TargetAura = {}
DrDamage.Consumables = {}
DrDamage.Calculation = {}

local function DrD_LoadAuras()
	local playerClass = select(2,UnitClass("player"))
	local playerHealer = (playerClass == "PRIEST") or (playerClass == "SHAMAN") or (playerClass == "PALADIN") or (playerClass == "DRUID")
	local playerCaster = (playerClass == "MAGE") or (playerClass == "PRIEST") or (playerClass == "WARLOCK")
	local playerMelee = (playerClass == "ROGUE") or (playerClass == "WARRIOR") or (playerClass == "HUNTER")
	local playerHybrid = (playerClass == "DEATHKNIGHT") or (playerClass == "DRUID") or (playerClass == "PALADIN") or (playerClass == "SHAMAN")
	local DK = (playerClass == "DEATHKNIGHT")
	local Aura = DrDamage.PlayerAura
	local GetSpellInfo = GetSpellInfo
	local horde = (UnitFactionGroup("player") == "Horde")
	
	--[[ NOTES: 
	 	School  = "Spells" means it applies to both heals and spells (in hybrid classes), not physical
	 	School = "Damage Spells" means it applies to all damaging spells (in hybrid classes), not physical
		School = "Healing" applies only to heals
		School = "Physical" applies only to physical type abilities, which is the default in melee module unless otherwise specified
		No school or spell applies to everything but healing
		Caster = true or Melee = true limits it to respective modules
	--]]
	
	if horde then
		--Bloodlust
		Aura[GetSpellInfo(2825)] = { School = "Spells", Mods = { ["castTime"] = function(v) return v/1.3 end }, Category = "30% haste", ID = 2825 }
	else
		--Heroism
		Aura[GetSpellInfo(32182)] = { School = "Spells", Mods = { ["castTime"] = function(v) return v/1.3 end }, Category = "30% haste", ID = 32182 }
		--Heroic Presence
		Aura[GetSpellInfo(28878)] = { Value = 1, ModType = "hitPerc", ID = 28878 }
	end
	
	if playerCaster or playerHybrid then
	--CORE	
		--Bloodgem Infusion
		Aura[GetSpellInfo(34379)] = { School = "Damage Spells", Value = 0.05, NoManual = true }
		--Vibrant Blood
		Aura[GetSpellInfo(35329)] = { School = "Damage Spells", Value = 0.1, NoManual = true }		
		--Hysteria (needs to be divided out from multiplier)
		Aura[GetSpellInfo(49016)] = { School = "Damage Spells", Value = -1 + 1/1.2, NoManual = true }		
		--Unrelenting Assault
		Aura[GetSpellInfo(64850)] = { School = "Spells", Value = -0.25, Ranks = 2, NoManual = true }
	--MODS
		--Totem of Wrath
		if not DK then
		Aura[GetSpellInfo(30706)] = { School = "Spells", Mods = { ["spellDmg"] = 280 }, ID = 30706 }
		end
		--Moonkin Aura
		Aura[GetSpellInfo(24907)] = { School = "Spells", Mods = { ["critPerc"] = 5 }, Category = "+5% crit", ID = 24907 }
		--Elemental Oath
		Aura[GetSpellInfo(53414)] = { School = "Spells", Mods = { ["critPerc"] = 5 }, Category = "+5% crit", ID = 53414 }
		--Wrath of Air
		Aura[GetSpellInfo(3738)] = { School = "Spells", Mods = { ["castTime"] = function(v) return v/1.05 end }, Caster = true, ID = 3738  }
		--Power Infusion
		Aura[GetSpellInfo(10060)] = { School = "Spells", Mods = { ["castTime"] = function(v) return v/1.2 end }, Caster = true, ID = 10060 }		
	end
	if playerHealer then
	--Buffs
		--Nether Portal - Serenity (Karazhan)
		Aura[GetSpellInfo(30422)] = { School = "Healing", Value = 0.05, Apps = 99, Caster = true, NoManual = true }
		--Luck of the Draw (Random LFG Bonus)
		Aura[(GetSpellInfo(72221) or "Luck of the Draw")] = { School = "Healing", Value = 0.05, Caster = true, NoManual = true }
	--Debuffs	
		--Stolen Soul (Exarch Maladaar)
		Aura[GetSpellInfo(32346)] = { School = "Healing", Value = -0.5, Caster = true, NoManual = true }		
		--Wretching Bile (Stratholme)
		Aura[GetSpellInfo(52527)] = { School = "Healing", Value = -0.35, Caster = true, NoManual = true }
		--Necrotic Strike (Icecrown)
		Aura[GetSpellInfo(60626)] = { School = "Healing", Value = -0.1, Apps = 1, Caster = true, NoManual = true }
		--Emerald Vigor (Valithria Dreamwalker)
		Aura[(GetSpellInfo(70873) or "Emerald Vigor")] = { School = "Healing", Value = 0.10, Apps = 1, Caster = true, NoManual = true }
	end
	if playerHealer or playerCaster or playerHybrid then
		--Shadow Crash ((Valithria Dreamwalker)
		Aura[GetSpellInfo(63277)] = { School = "Spells", ModType = "Special", NoManual = true }
		DrDamage.Calculation[GetSpellInfo(63277)] = function( calculation )
			if calculation.healingSpell then
				calculation.dmgM = calculation.dmgM * 0.25
			else
				calculation.dmgM = calculation.dmgM * 2
			end
		end
		
	end
	
	--Target
	Aura = DrDamage.TargetAura

	--Buffs
	--Icebound Fortitude (TODO: How to implement this reduction properly?)
	Aura[GetSpellInfo(48792)] = { Value = -0.2, NoManual = true }
	--Bone Shield
	Aura[GetSpellInfo(49222)] = { Value = -0.2, NoManual = true }
	--Blade Barrier
	Aura[GetSpellInfo(51789)] = { Value = -0.01, Ranks = 5, NoManual = true }
	--Totem of Wrath
	Aura[GetSpellInfo(57722)] = { Value = 3, ModType = "critPerc", Category = "+3% crit", ID = 57722 }
	--Heart of the Crusader
	Aura[GetSpellInfo(20335)] = { Value = 1, Ranks = 3, ModType = "critPerc", Category = "+3% crit", ID = 20335 }
	--Master Poisoner
	Aura[GetSpellInfo(58410)] = { Value = 1, Ranks = 3, ModType = "critPerc", Category = "+3% crit", ID = 58410 }
	--Focused Will
	Aura[GetSpellInfo(45234)] = { ModType = "Special", NoManual = true }
	local UnitExists = UnitExists
	local UnitIsFriend = UnitIsFriend
	DrDamage.Calculation[GetSpellInfo(45234)] = function( calculation, _, _, _, apps, _, rank )
		rank = rank and tonumber(string_match(rank,"%d+")) or 3
		apps = apps or 3
		if calculation.healingSpell then
			if not UnitExists("target") or UnitIsFriend("target","player") then
				calculation.dmgM = calculation.dmgM * (1 + (rank * 0.01 + 0.02) * apps)
			end
		elseif UnitExists("target") and not UnitIsFriend("target","player") then
			calculation.dmgM = calculation.dmgM * (1 - (rank * 0.01 + 0.01) * apps)
		end	
	end	
	
	if playerCaster or playerHybrid then
	--General
		--Curse of the Elements (Warlock)
		Aura[GetSpellInfo(1490)] = { School = "Damage Spells", Ranks = 5, Value = { 0.06, 0.08, 0.1, 0.11, 0.13 }, Category = "+13% dmg", ID = 1490 }
		--Earth and Moon (Druid)
		Aura[GetSpellInfo(60431)] = { School = "Damage Spells", Ranks = 3, Value = { 0.04, 0.09, 0.13 }, Category = "+13% dmg", ID = 60431 }
		--Ebon Plague (Death Knight)
		Aura[GetSpellInfo(51735)] = { School = "Damage Spells", Ranks = 3, Value = { 0.04, 0.09, 0.13 }, Category = "+13% dmg", ID = 51735 }	
		--Shadow Mastery (Improved Shadow Bolt)
		Aura[GetSpellInfo(17800)] = { School = "Damage Spells", Value = 5, ModType = "critPerc", Category = "+5% target crit", ID = 17800 }
		--Improved Scorch (Mage)
		Aura[GetSpellInfo(22959)] = { School = "Damage Spells", Value = 5, ModType = "critPerc", Category = "+5% target crit", ID = 22959 }
		--Winter's Chill (Mage)
		Aura[GetSpellInfo(12579)] = { School = "Damage Spells", Value = 5, ModType = "critPerc", Category = "+5% target crit", ID = 12579 }	
		--Misery (Priest)
		Aura[GetSpellInfo(33196)] = { School = "Damage Spells", Value = 1, Ranks = 3, ModType = "hitPerc", Category = "+3% hit", ID = 33196 }
		--Anti-Magic Shell
		Aura[GetSpellInfo(48707)] = { School = "Damage Spells", Value = -0.75, NoManual = true }
	--Weapon Debuffs
		--Spell Vulnerability (from Nightfall)
		Aura[GetSpellInfo(23605)] = { School = "Damage Spells", Value = 0.15, NoManual = true }
	end
	--Buffs
	if playerHealer then	
		--Amplify Magic
		Aura[GetSpellInfo(1008)] = { School = "Healing", ModType = "spellDmg", Ranks = 7, Value = { 16, 32, 53, 80, 96, 128, 255 }, ID = 1008 }
		--Demon Armor
		Aura[GetSpellInfo(27260)] = { School = "Healing", Value = 0.2, ID = 27260 }
		--Vampiric Blood
		Aura[GetSpellInfo(55233)] =  { School = "Healing", Value = 0.35, NoManual = true }
		--Divine Hymn
		Aura[GetSpellInfo(64844)] =  { School = "Healing", Value = 0.1, NoManual = true }
		--Guardian Spirit
		Aura[GetSpellInfo(47788)] =  { School = "Healing", Value = 0.4, NoManual = true }
		--Tree of Life
		Aura[GetSpellInfo(34123)] =  { School = "Healing", ModType = "Special", Category = "+6% Healing", ID = 34123 }
		local passive = select(2,GetSpellInfo(5420))
		DrDamage.Calculation[GetSpellInfo(34123)] = function( calculation, _, _, _, _, _, rank )
			if rank == passive then
				calculation.dmgM = calculation.dmgM * 1.06
			end
		end
		--Tenacity
		Aura[GetSpellInfo(58549)] = { School = "Healing", Texture = "Interface\\Icons\\Ability_Warrior_StrengthOfArms", Apps = 1, Value = 0.18, NoManual = true }
	--Debuffs
	--Player
		--Mortal Strike
		Aura[GetSpellInfo(12294)] = { School = "Healing", Value = -0.5, Caster = true, Category = "Mortal Strike", Manual = GetSpellInfo(12294), ID = 12294 }
		--Aimed Shot
		Aura[GetSpellInfo(49050)] = Aura[GetSpellInfo(12294)]
		--Wound Poison I-VII
		Aura[GetSpellInfo(13218)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(13222)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(13223)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(13224)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(27189)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(57974)] = Aura[GetSpellInfo(12294)]
		Aura[GetSpellInfo(57975)] = Aura[GetSpellInfo(12294)]
		--Furious Attacks (Fury Warrior talent)
		Aura[GetSpellInfo(56112)] = { School = "Healing", Value = -0.25, Apps = 2, Caster = true, Category = "Mortal Strike", Manual = GetSpellInfo(12294), ID = 12294 }		
		--Shadow Embrace
		Aura[GetSpellInfo(32385)] = { School = "Healing", ModType = "dmgM_dot", Value = -0.01, Ranks = 5, Apps = 2, Multiply = true, Caster = true, ID = 32385 }
		--Mind Trauma
		Aura[GetSpellInfo(48301)] = { School = "Healing", Value = -0.2, Caster = true, Category = "Mortal Strike", ID = 48301 }			
	--NPC
		--25% reduction
		--Fetid Rot (King Ymiron)
		Aura[GetSpellInfo(48291)] = { School = "Healing", Value = -0.25, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Wounding Strike (Chrono-Lord Epoch - Stratholme)
		Aura[GetSpellInfo(52771)] = Aura[GetSpellInfo(48291)]
		--Dark Volley (Ulduar - Guardian of Yogg-Saron)
		Aura[GetSpellInfo(63038)] = Aura[GetSpellInfo(48291)]
		--Hopelessness
		Aura[GetSpellInfo(72390)] = Aura[GetSpellInfo(48291)]
		
		--50% reduction (Mortal Strike)
		--Mortal Cleave (random mobs)
		Aura[GetSpellInfo(22859)] = { School = "Healing", Value = -0.5, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Mutated Infection (Rotface)
		Aura[(GetSpellInfo(69674) or "Mutated Infection")] = Aura[GetSpellInfo(22859)]
		--Soul Strike (Halls of Lightning, Mana-Tombs)
		Aura[GetSpellInfo(32315)] = Aura[GetSpellInfo(22859)]
		--Curse of the Deadwood (Felwood)
		Aura[GetSpellInfo(13583)] = Aura[GetSpellInfo(22859)]
		--Mortal Cleave (random mobs)
		Aura[GetSpellInfo(22859)] = Aura[GetSpellInfo(22859)]
		--Arcing Smash (Gurtogg)
		Aura[GetSpellInfo(40599)] = Aura[GetSpellInfo(22859)]
		
		--75% reduction
		--Veil of Shadow (Multiple places)
		Aura[GetSpellInfo(17820)] = { School = "Healing", Value = -0.75, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Veil of Shadow Alternate (Different Localized name)
		Aura[GetSpellInfo(69633)] = Aura[GetSpellInfo(17820)]
		--Gehennas' Curse
		Aura[GetSpellInfo(19716)] = Aura[GetSpellInfo(17820)]
		
		--100% reduction
		--Necrotic Aura (Loatheb)
		Aura[GetSpellInfo(55593)] = { School = "Healing", Value = -1, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Embrace of the Vampyr (Prince Taldaram)
		Aura[GetSpellInfo(59513)] = Aura[GetSpellInfo(55593)]
		--Enfeeble (Prince Malchezaar)
		Aura[GetSpellInfo(30843)] = Aura[GetSpellInfo(55593)]
		
		--Others
		--Nether Portal - Dominance (Kharazan)
		Aura[GetSpellInfo(30423)] = { School = "Healing", Value = -0.01, Apps = 10, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Dark Touched (Eredar Twins)
		Aura[GetSpellInfo(45347)] = { School = "Healing", Value = -0.05, Apps = 1, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Mortal Wound (Naxxramas etc)
		Aura[GetSpellInfo(28467)] = { School = "Healing", Value = -0.1,  Apps = 1, Caster = true, Category = "Mortal Strike", NoManual = true }
		--Chop (random 70-80 mobs)
		Aura[GetSpellInfo(43410)] = { School = "Healing", Value = -0.1,  Caster = true, Category = "Mortal Strike", NoManual = true }
		--Suppression (Valithria Dreamwalker)
		Aura[(GetSpellInfo(70588) or "Suppression")] = { School = "Healing", Value = -0.1, Caster = true, Category = "Mortal Strike", NoManual = true }		
		--Shroud of Darkness (Zuramat the Obliterator)
		Aura[GetSpellInfo(54525)] = { School = "Healing", Value = -0.2, Apps = 1, Caster = true, Category = "Mortal Strike", NoManual = true }
	end
	if playerMelee or playerHybrid then
	--Buffs
	--Player
		--Savage Combat
		Aura[GetSpellInfo(58683)] = { School = "Physical", Value = 0.02, Ranks = 2, Category = "+4% Physical", ID = 58683 }
		--Blood Frenzy
		Aura[GetSpellInfo(30070)] = { School = "Physical", Value = 0.02, Ranks = 2, Category = "+4% Physical", ID = 30070 }
		--Hemorrhage
		Aura[GetSpellInfo(16511)] = { School = "Physical", Value = { 13, 21, 29, 42, 75 }, Ranks = 5, ModType = "finalMod", ActiveAura = "Hemorrhage", ID = 16511 }
		--Trauma
		Aura[GetSpellInfo(46856)] = { School = "Physical", Value = 0.15, Ranks = 2, ModType = "bleedBonus", Category = "+30% bleed", Manual = GetSpellInfo(33917), ID = 33917 }
		--Mangle (Bear)
		Aura[GetSpellInfo(33878)] = { School = "Physical", Value = 0.3, ModType = "bleedBonus", Category = "+30% bleed", Manual = GetSpellInfo(33917), ID = 33917 }
		--Mangle (Cat)
		Aura[GetSpellInfo(33876)] = Aura[GetSpellInfo(33878)]
		--Stampede
		Aura[GetSpellInfo(57393)] = { School = "Physical", Value = 0.25, ModType = "bleedBonus", Category = "+30% bleed", NoManual = true }
	--Special	
		--Armor Disruption
		Aura[GetSpellInfo(36482)] = { School = "Physical", Value = 0.05, Apps = 5, NoManual = true }		
	--Debuffs
	--Player
		--Shadowform
		Aura[GetSpellInfo(15473)] = { School = "Physical", Value = -0.15, NoManual = true }
		--Sunder Armor
		Aura[GetSpellInfo(7386)] = { School = "Physical", Apps = 5, Value = 0.04, ModType = "armorM", Category = "-20% Armor", Manual = GetSpellInfo(7386), ID = 7386 }
		--Expose Armor
		Aura[GetSpellInfo(8647)] = { School = "Physical", Value = 0.2, ModType = "armorM", Category = "-20% Armor", Manual = GetSpellInfo(7386), ID = 7386 }
		--Acid Spit
		Aura[GetSpellInfo(55749)] = { School = "Physical", Apps = 2, Value = 0.1, ModType = "armorM", Category = "-20% Armor", Manual = GetSpellInfo(7386), ID = 7386 }
		--Faerie Fire
		Aura[GetSpellInfo(770)] = { School = "Physical", Value = 0.05, ModType = "armorM", Category = "-5% Armor", Manual = GetSpellInfo(770), ID = 770 }
		--Faerie Fire (Feral)
		Aura[GetSpellInfo(16857)] = Aura[GetSpellInfo(770)]
		--Sting
		Aura[GetSpellInfo(56626)] = Aura[GetSpellInfo(770)]
		--Curse of Weakness
		Aura[GetSpellInfo(702)] = Aura[GetSpellInfo(770)]
		
	--NPC
		--Holyform
		Aura[GetSpellInfo(46565)] = { School = "Physical", Value = -0.2, NoManual = true }
	end
	
	local Consumables = DrDamage.Consumables
	if (playerCaster or playerHybrid or playerHealer) and not DK then
		--Flask of the Frost Wyrm
		Consumables[GetSpellInfo(53755)] = { School = "Spells", Mods = { ["spellDmg"] = 125 }, Category = "Battle Elixir", Category2 = "Guardian Elixir" }
		--Spellpower Elixir
		Consumables[GetSpellInfo(33721)] = { School = "Spells", Mods = { ["spellDmg"] = 58 }, Category = "Battle Elixir" }
		--Flask of Pure Mojo
		Consumables[GetSpellInfo(54212)] = { School = "Spells", Mods = { ["manaRegen"] = 7.6 }, Caster = true, Category = "Battle Elixir", Category2 = "Guardian Elixir" }
		--Elixir of Mighty Mageblood
		Consumables[GetItemInfo(40109) or "Elixir of Mighty Mageblood"] = { School = "Spells", Mods = { ["manaRegen"] = 4.8 }, Alt = GetSpellInfo(53764), Caster = true, Category = "Guardian Elixir" }
		--Food
		Consumables[L["+46 Spell Power Food"]] = { School = "Spells", Mods = { ["spellDmg"] = 46, }, Alt = GetSpellInfo(57327), Category = "Food" }
	end
	if playerMelee or playerHybrid then
		--Flask of Endless Rage
		Consumables[GetSpellInfo(53760)] = { Mods = { ["AP"] = 180 }, Category = "Battle Elixir", Category2 = "Guardian Elixir" }
		--Elixir of Wrath
		Consumables[GetSpellInfo(53746)] = { Mods = { ["AP"] = 90 }, Category = "Battle Elixir" }
		--Elixir of Demonslaying
		Consumables[GetSpellInfo(11406)] = { Mods = { ["AP"] = 265 }, Category = "Battle Elixir" }
		--Food
		Consumables[L["+60 AP Food"]] = { Mods = { ["AP"] = 60 }, Alt = GetSpellInfo(57079), Category = "Food" }
		Consumables[L["+80 AP Food"]] = { Mods = { ["AP"] = 80 }, Alt = GetSpellInfo(57079), Category = "Food" }
	end
	
	--TODO: Fix hit rating to hit for casters
	--Elixir of Accuracy
	Consumables[GetItemInfo(44325) or "Elixir of Accuracy"] = { School = { ["Damage Spells"] = true, ["Ranged"] = true, ["Physical"] = true }, Mods = { ["hitPerc"] = 1.372 }, Alt = GetSpellInfo(60340), Category = "Battle Elixir" }
	--Elixir of Deadly Strikes
	Consumables[GetItemInfo(44327) or "Elixir of Deadly Strikes"] = { School = { ["Spells"] = true, ["Ranged"] = true, ["Physical"] = true }, Mods = { ["critPerc"] = 0.9802, }, Alt = GetSpellInfo(60341), Category = "Battle Elixir" }
	--Food
	Consumables[L["+40 Hit Rating Food"]] = { School = { ["Damage Spells"] = true, ["Ranged"] = true, ["Physical"] = true }, Mods = { ["hitPerc"] = 1.22 }, Alt = GetSpellInfo(33263), Category = "Food" }
	Consumables[L["+40 Critical Strike Rating Food"]] = { School = { ["Spells"] = true, ["Ranged"] = true, ["Physical"] = true }, Mods = { ["critPerc"] = 0.871 }, Alt = GetSpellInfo(33263), Category = "Food" }
end

DrD_LoadAuras()
DrD_LoadAuras = nil