local _, playerClass = UnitClass("player")
local playerHealer = (playerClass == "PRIEST") or (playerClass == "SHAMAN") or (playerClass == "PALADIN") or (playerClass == "DRUID")
local playerCaster = (playerClass == "MAGE") or (playerClass == "PRIEST") or (playerClass == "WARLOCK")
local playerMelee = (playerClass == "ROGUE") or (playerClass == "WARRIOR") or (playerClass == "HUNTER")
local playerHybrid = (playerClass == "DEATHKNIGHT") or (playerClass == "DRUID") or (playerClass == "PALADIN") or (playerClass == "SHAMAN")

--Libraries
local L = LibStub("AceLocale-3.0"):GetLocale("DrDamage", true)
local ACD = LibStub("AceConfigDialog-3.0")
local GT = LibStub:GetLibrary("LibGratuity-3.0")
local LSM = LibStub:GetLibrary("LibSharedMedia-3.0", true)
local DrDamage = DrDamage

--General
local settings
local _G = getfenv(0)
local type = type
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local math_floor = math.floor
local math_min = math.min
local math_max = math.max
local string_match = string.match
local string_format = string.format
local string_find = string.find
local string_sub = string.sub
local string_gsub = string.gsub
local string_len = string.len
local select = select
local next = next

--Module
local GameTooltip = GameTooltip
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitLevel = UnitLevel
local UnitDamage = UnitDamage
local UnitPower = UnitPower
local UnitPowerMax = UnitPowerMax
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitAttackPower = UnitAttackPower
local UnitIsUnit = UnitIsUnit
local UnitClassification = UnitClassification
local UnitExists = UnitExists
local UnitIsPlayer = UnitIsPlayer
local BOOKTYPE_SPELL = BOOKTYPE_SPELL
local GetSpellName = GetSpellName
local GetSpellInfo = GetSpellInfo
local GetMacroSpell = GetMacroSpell
local GetActionInfo = GetActionInfo
local GetCursorInfo = GetCursorInfo
local GetInventoryItemLink = GetInventoryItemLink
local GetItemInfo = GetItemInfo
local GetItemGem = GetItemGem
local GetTime = GetTime
local GetWeaponEnchantInfo = GetWeaponEnchantInfo
local GetTalentInfo = GetTalentInfo
local GetNumTalentTabs = GetNumTalentTabs
local GetNumTalents = GetNumTalents
local GetGlyphSocketInfo = GetGlyphSocketInfo
local GetCombatRatingBonus = GetCombatRatingBonus
local HasAction = HasAction
local IsEquippedItem = IsEquippedItem
local IsAddOnLoaded = IsAddOnLoaded
local IsAltKeyDown = IsAltKeyDown
local IsControlKeyDown = IsControlKeyDown
local IsShiftKeyDown = IsShiftKeyDown
local SecureButton_GetModifiedAttribute = SecureButton_GetModifiedAttribute
local SecureButton_GetEffectiveButton = SecureButton_GetEffectiveButton
local ActionButton_GetPagedID = ActionButton_GetPagedID
local InCombatLockdown = InCombatLockdown

--Module variables
local playerCompatible, DrD_Font
local spellInfo, talentInfo, talents, PlayerHealth, TargetHealth
local updateSetItems
local dmgMod
local ModStateEvent

DrDamage.visualChange = true
DrDamage.fullUpdate = false

--Local functions
function DrDamage.ClearTable(table)
	for k in pairs( table ) do
		table[k] = nil
	end
end
function DrDamage.Round(x, y)
	return math_floor( x * 10 ^ y + 0.5 ) / 10 ^ y
end
function DrDamage.MatchData( data, ... )
	if not data or not ... then
		return false
	end

	if type( data ) == "table" then
		for i = 1, select('#', ...) do
			if data[1] then
				for _, dataName in ipairs( data ) do
					for i = 1, select('#', ...) do
						if dataName == select(i, ...) then
							return true
						end
					end
				end
			else
				if data[select(i, ...)] then
					return true
				end
			end
		end
	else	
		for i = 1, select('#', ...) do
			if data == select(i, ...) then
				return true
			end
		end
	end
	
	return false
end
local function DrD_Set(n)
	return function(info, v)
		settings[n] = v
		if not DrDamage.fullUpdate then DrDamage.fullUpdate = DrDamage:ScheduleTimer("UpdateAB", 0.4) end			
	end
end

local DrD_ClearTable = DrDamage.ClearTable
local DrD_Round = DrDamage.Round
local DrD_MatchData = DrDamage.MatchData
local ABfunc, ABdisable
local ABtable, ABobjects, ABrefresh = {}, {}, {}

local function DrD_DetermineAB()
	--Default
	for i = 1, 6 do
		for j = 1, 12 do
			table.insert(ABobjects,_G[((select(i,"ActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton", "MultiBarRightButton", "MultiBarLeftButton", "BonusActionButton"))..j)])
		end
	end
	if playerClass == "SHAMAN" then
		for i = 1, 4 do
			table.insert(ABobjects,_G[("MultiCastActionButton"..i)])
		end
	end
	if playerClass == "PALADIN" then
		local func = function(button)
			if button then
				local _, name = GetShapeshiftFormInfo(button:GetID()) 
				local _, rank = GetSpellInfo(name)
				return nil, name, rank
			end
		end
		ABrefresh["ShapeshiftButton"] = function()
			for i=1,7 do
				ABtable["ShapeshiftButton"..i] = func
			end
		end
	end
	if IsAddOnLoaded("Bartender4") then
		local func = function(button) if button.Secure then return button.Secure:GetActionID() else return button:GetActionID() end end
		ABrefresh["BT4Button"] = function()
			for i=1,120 do
				ABtable["BT4Button"..i] = func
			end
		end
		ABdisable = true
	end
	if nUI_DrDamageIntegration then
		ABrefresh["nUI"] = nUI_DrDamageIntegration( ABtable )
		ABdisable = true
	end
	if IsAddOnLoaded("Nurfed") then
		local func = function(button)
			if button.spell then
				if button.type == "spell" then
					local pid = Nurfed:getspell(button.spell)
					if pid then
						return nil, GetSpellName(pid, BOOKTYPE_SPELL)
					end
				elseif button.type == "macro" then
					local action, rank = GetMacroSpell(button.spell)
					if action then 							
						return nil, action, rank
					end
				end
			end
		end	
		ABrefresh["Nurfed_Button"] = function()
			for i=1,120 do
				ABtable["Nurfed_Button"..i] = func
			end
		end
		ABdisable = true
	end	
	if IsAddOnLoaded("Macaroon") then
		local func = function(button)
			if button.config and button.config.type == "action" then
				return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button))
			end
			if button.macroshow then
				return nil, string_match(button.macroshow,"[^%(]+"), button.macrorank
			end
			if button.macrospell then
				return nil, string_match(button.macrospell,"[^%(]+"), button.macrorank
			end
		end
		ABrefresh["MacaroonButton"] = function()
			for _, button in pairs(Macaroon.Buttons) do
				ABtable[button[1]:GetName()] = func
			end
		end
	end	
	if IsAddOnLoaded("Dominos") then
		local func = function(button) return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button)) end
		ABrefresh["DominosActionButton"] = function()
			for i=1,120 do
				ABtable["DominosActionButton"..i] = func
			end
		end
	end
	if IsAddOnLoaded("IPopBar") then
		local func = function(button) return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button)) end
		ABrefresh["IPopBarButton"] = function()
			for i=1,120 do
				ABtable["IPopBarButton"..i] = func
			end
		end
	end
	if IsAddOnLoaded("CT_BarMod") then
		--There's an object list available CT_BarMod.actionButtonList, index = actionID, value.hasAction
		local func = function(button) return button.object.id end
		ABrefresh["CT_BarModActionButton"] = function()
			for i=13,120 do
				ABtable["CT_BarModActionButton"..i] = func
			end		
		end
	end
	if IsAddOnLoaded("RDX") then
		local func = function(button) return button:GetAttribute("action") end
		ABrefresh["RDX_ActionBars"] = function()
			for i=1,120 do
				ABtable["VFLButton"..i] = func
			end
		end
	end
	if not next(ABrefresh) then
		ABrefresh, ABtable = nil, nil
	end
	DrDamage:RefreshAB()
end

function DrDamage:RefreshAB()
	if ABrefresh then
		for k in pairs(ABtable) do
			ABtable[k] = nil
		end
		for _, func in pairs(ABrefresh) do
			func()
		end
	end
end

--Options table
DrDamage.options = { type='group', args = {} }
--Defaults table
DrDamage.defaults = {
	profile = {
		--General
		FullText = false,
		--Actionbar
		DisplayType_M = "AvgTotal",
		DisplayType = "AvgTotal",
		CastsLeft = false,
		CastsLeftDmg = false,
		--
		ABText = true,
		DefaultAB = true,
		Font = GameFontNormal:GetFont(),
		FontEffect = "OUTLINE",
		FontSize = 10,
		FontXPosition = 0,
		FontYPosition = 0,
		FontColorDmg = { r = 1, g = 1, b = 0.2 },
		FontColorHeal = { r = 0.4, g = 1.0, b = 0.3 },
		FontColorMana = { r = 0.4, g = 0.8, b = 1.0 },
		UpdateShift = false,
		UpdateAlt = false,
		UpdateCtrl = false,
		Tooltip = "Always",
		DefaultColor = false,
		Hints = true,
		SwapCalc = false,
		--Tooltip:
		PlusDmg = false,
		Coeffs = true,
		DispCrit = true,
		DispHit = true,
		AvgHit = true,
		AvgCrit = true,
		Ticks = true,
		Total = true,
		Extra = true,
		Next = true,
		DPS = true,
		DPM = true,
		DPP = true,
		Doom = true,
		Casts = true,
		ManaUsage = false,	
		--Comparisons
		CompareDmg = true,
		CompareHit = false,
		CompareCrit = false,
		CompareHaste = false,
		--Custom Stats
		Custom = false,
		CustomAdd = true,	
		SpellDamage = 0,
		AP = 0,
		HitRating = 0,
		CritRating = 0,
		HasteRating = 0,
		MP5 = 0,
		--Calculation
		HitCalc = false,
		CritDepression = false,
		TwoRoll = false,
		TwoRoll_M = false,
		ManaConsumables = false,
		Dodge = false,
		Parry = false,
		Glancing = false,
		TargetAmount = 1,
		ComboPoints = 0,
		HitTarget = false,
		BossArmor = false,
		TargetLevel = 3,		
		Armor = 0,
		Resilience = 0,
		--Buffs
		PlayerAura = {},
		TargetAura = {},
		Consumables = {},
	}
}

function DrDamage:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("DrDamageDB", DrDamage.defaults)
	--self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	settings = self.db.profile
	LibStub("AceConfig-3.0"):RegisterOptionsTable("DrDamage", DrDamage.options)
	ACD:AddToBlizOptions("DrDamage")
	self:RegisterChatCommand("drd", function() ACD:Open("DrDamage") end)
	self:RegisterChatCommand("drdmg", function() ACD:Open("DrDamage") end)
	self:RegisterChatCommand("drdamage", function() ACD:Open("DrDamage") end)
	local LDB = LibStub:GetLibrary("LibDataBroker-1.1", true)
	if LDB then
		LDB:NewDataObject("DrDamage", {
			type = "launcher",
			icon = "Interface\\Icons\\Spell_Holy_SearingLightPriest",
			OnClick = function(clickedframe, button)
				ACD:Open("DrDamage")
			end,
		})
	end
end

function DrDamage:OnEnable()
	if self.PlayerData then
		self.ClassSpecials = {}
		self.DmgCalculation = {}
		self.SetBonuses = {}
		self.RelicSlot = {}
		self.talents = {}
		self:PlayerData()
		self.PlayerData = nil
		PlayerHealth = self.PlayerHealth
		TargetHealth = self.TargetHealth
		spellInfo = self.spellInfo
		talentInfo = self.talentInfo
		talents = self.talents
		self:CommonData()
		self.CommonData = nil		
	elseif not self.spellInfo and not self.talentInfo then
		return
	end
	
	self:RefreshConfig()
  	self:RegisterEvent("PLAYER_ENTERING_WORLD", function() self:UnregisterEvent("PLAYER_ENTERING_WORLD"); self:ScheduleTimer("Load", 3) end)
end

function DrDamage:OnDisable()
	if playerCompatible then
		local tempType = settings.ABText
		settings.ABText = false
		self:UpdateAB()
		settings.ABText = tempType
	end	
end

function DrDamage:Load()
	if self.GeneralOptions then
		self:GeneralOptions()
		self.GeneralOptions = nil
	end
	if self.Caster_OnEnable then
		self:Caster_OnEnable()
		self.Caster_OnEnable = nil
	end
	if self.Melee_OnEnable then
		self:Melee_OnEnable()
		self.Melee_OnEnable = nil
	end
	if DrD_DetermineAB then 
		DrD_DetermineAB()
		DrD_DetermineAB = nil
	end
	if not playerCompatible then
		--Make talent spell tables better suited for iteration
		for k in pairs( talentInfo ) do
			local talent = talentInfo[k]
			for i=1,#talent do
				local spelltable = talent[i].Spells
				if type(spelltable) == "table" then
					for j,v in ipairs( spelltable ) do
						spelltable[v] = true
						spelltable[j] = nil
					end
				end
			end
		end
		--Create talent tables
		for k, v in pairs( spellInfo ) do
			v["Talents"] = {}
			if v["Secondary"] then
				v["Secondary"]["Talents"] = {}
			end
		end
	end
	if (settings.Font ~= self.defaults.profile.Font) and not CreateFrame("Frame"):CreateFontString("DrDamage-Font"):SetFont(settings.Font,10) then
		settings.Font = GameFontNormal:GetFont()
	end
	DrD_Font = DrD_Font or settings.Font
	self:MetaGems()
	self:UpdateGlyphs()
	self:UpdateTalents()
	playerCompatible = true

	if LSM then
		LSM.RegisterCallback(self, "LibSharedMedia_Registered", function(event, media) 
			if media == "font" then
				self.options.args.General.args.Font.values = LSM:HashTable("font")
			end
		end)
		LSM.RegisterCallback(self, "LibSharedMedia_SetGlobal", function(event, media, key) 
			if media == "font" then
				DrD_Font = LSM:Fetch("font")
				self.visualChange = true
				self:UpdateAB()
			end
		end)
	end	
	
	self:SecureHook(GameTooltip, "SetAction")
	self:SecureHook(GameTooltip, "SetSpell")
  	self:RegisterEvent("CHARACTER_POINTS_CHANGED","UPDATE_TALENTS")
  	self:RegisterEvent("PLAYER_TALENT_UPDATE","UPDATE_TALENTS")
	self:RegisterEvent("GLYPH_ADDED", "UpdateGlyphs")
	self:RegisterEvent("GLYPH_UPDATED", "UpdateGlyphs")
	self:RegisterEvent("GLYPH_REMOVED", "UpdateGlyphs")
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig", true)
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig", true)
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig", true, true)
	
	if playerClass == "PALADIN" then
		self:SecureHook(GameTooltip, "SetShapeshift")
	end	
		
	if settings.ABText then
		if settings.UpdateAlt or settings.UpdateCtrl or settings.UpdateShift then
			self:RegisterEvent( "MODIFIER_STATE_CHANGED" )
			ModStateEvent = true
		end
		if PlayerHealth or TargetHealth then
			self:RegisterBucketEvent("UNIT_HEALTH", 2.5)
		end
		if (playerClass == "SHAMAN" or playerClass == "WARLOCK" or playerClass == "ROGUE" or playerClass == "DEATHKNIGHT") then
			self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", 1.5)
		end
		if (playerClass == "SHAMAN") then
			self:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
		end
		self:RegisterEvent("ACTIONBAR_PAGE_CHANGED")
		self:RegisterEvent("ACTIONBAR_HIDEGRID")
		self:RegisterEvent("EXECUTE_CHAT_LINE")
		self:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MACROS")
		self:RegisterEvent("LEARNED_SPELL_IN_TAB")
		self:RegisterEvent("UNIT_AURA")
	end
end

function DrDamage:RefreshConfig(cb, reset)
	if self.Caster_RefreshConfig then self:Caster_RefreshConfig() end
	if self.Melee_RefreshConfig then self:Melee_RefreshConfig() end
	settings = self.db.profile
	if reset then
		DrD_Font = DrDamage.defaults.profile.Font
	end	
	if cb then
		self.visualChange = true
		self:UpdateAB()
	end
end

function DrDamage:CommonData()
	if select(2, UnitRace("player")) == "Draenei" then
		if playerClass == "PALADIN" or playerClass == "PRIEST" or playerClass == "SHAMAN" or playerClass == "MAGE" or playerClass == "DEATHKNIGHT" then
			--Gift of the Naaru
			spellInfo[GetSpellInfo(28880)] = {
						["Name"] = "Gift of the Naaru",
						[0] = { School = { "Holy", "Healing" }, Cooldown = 180, bonusFactor = 1, eDot = true, eDuration = 15, sTicks = 3, NoSchoolTalents = true },
						[1] = { 35, 35, },
			}
			self.Calculation["Gift of the Naaru"] = function( calculation )
				local b, p, n = UnitAttackPower("player")
				local ap = b + p + n
				if ap > calculation.spellDmg then 
					calculation.spellDmg = ap
					calculation.spellDmgM = 1.1
				end
				calculation.minDam = calculation.minDam + UnitLevel("player") * 15
				calculation.maxDam = calculation.minDam
			end
		elseif playerClass == "WARRIOR" or playerClass == "HUNTER" then
			self.ClassSpecials[GetSpellInfo(28880)] = function()
				local amount = 35 + UnitLevel("player") * 15 + 1.1 * self:GetAP()
				return amount, true
			end
		end
	end
	if select(2, UnitRace("player")) == "BloodElf" then
		if playerClass == "PALADIN" or playerClass == "PRIEST" or playerClass == "MAGE" or playerClass == "HUNTER" or playerClass == "WARLOCK" then
			--Arcane Torrent
			self.ClassSpecials[GetSpellInfo(28730)] = function()
				return 0.06 * UnitPowerMax("player",0), false, true
			end
		end
	end
	--Lifeblood
	if not playerMelee then
			spellInfo[GetSpellInfo(55428)] = {
						["Name"] = "Lifeblood",
						[0] = { School = { "Nature", "Healing" }, Cooldown = 180, bonusFactor = 0, eDot = true, eDuration = 5, sTicks = 1, SelfHeal = true, NoSchoolTalents = true, NoDPM = true, NoCasts = true },
						[1] = { 300, 300, },
						[2] = { 480, 480, },
						[3] = { 720, 720, },
						[4] = { 900, 900, },
						[5] = { 1200, 1200, },
						[6] = { 3600, 3600, },
			}
			self.Calculation["Lifeblood"] = function( calculation, _, Talents )
				calculation.minDam = calculation.minDam + 0.016 * UnitHealthMax("player")
				calculation.maxDam = calculation.minDam
				if Talents["Lifeblood Bonus"] then
					if playerClass == "DRUID" and GetShapeshiftForm() ~= 3 then return end
					calculation.dmgM = calculation.dmgM * (1 + Talents["Lifeblood Bonus"])
				end
			end
	else
		self.ClassSpecials[GetSpellInfo(55428)] = function(rank)
			local base = select(rank or 6, 300, 480, 720, 900, 1200, 3600) + 0.016 * UnitHealthMax("player")
			local mult = 1
			if playerClass == "ROGUE" then
				--Quick Recovery
				mult = mult * (1 + select((self.talents[GetSpellInfo(31244)] or 3),0.1,0.2,0))
			end
			return (mult * base), true
		end
	end
end

function DrDamage:GeneralOptions()
	self.options.args.General = {
		type = 'group', 
		name = L["General"],
		desc = L["General options."], 
		order = 1,
		args = {		
			Reset = {
				type = 'execute',
				name = L["Reset Options"],
				confirm = true,
				order = 2,
				func = function() self.db:ResetProfile() end,
			},		
			Actionbar = {
				type = 'group',
				name = L["Actionbar"],
				order = 50,
				args = {
					--[[
					MeleeTitle = {
						type = 'header',
						name = L["Melee"] .. "/" .. L["Ranged"], 
						--desc = L["Melee options."], 
						order = 1,
						hidden = playerCaster,
					},
					--]]
					DisplayType = {
						type = 'select',
						name = L["Melee"] .. "/" .. L["Ranged"],
						desc = L["Choose what to display on the actionbar."],
						order = 2,
						hidden = playerCaster,
						values =  { 
							["AvgTotal"] = L["AvgTotal"], 
							["Avg"] = L["Avg"], 
							["Min"] = L["Min"], 
							["Max"] = L["Max"], 
							["AvgCrit"] = L["AvgCrit"], 
							["MinCrit"] = L["MinCrit"], 
							["MaxCrit"] = L["MaxCrit"], 
							["DPM"] = L["DPM"], 
							["PowerCost"] = L["PowerCost"],
						},
						get =  function() return settings["DisplayType_M"] end,
						set =  function(info,v) settings["DisplayType_M"] = v; self:UpdateAB() end,
					},
					--[[
					CasterTitle = {
						type = 'header',
						name = L["Spells"], 
						--desc = L["Caster options."],
						order = 3,
						hidden = playerMelee,
					},
					--]]
					DisplayType2 = {
						type = 'select',
						name = L["Spells"],
						desc = L["Choose what to display on the actionbar."],
						order = 4,
						hidden = playerMelee,
						values =  {
							["Avg"] = L["Avg"],
							["AvgTotal"] = L["AvgTotal"],
							["AvgHit"] = L["AvgHit"],
							["AvgHitTotal"] = L["AvgHitTotal"], 
							["MinHit"] = L["MinHit"], 
							["MaxHit"] = L["MaxHit"], 
							["Max"] = L["Max"], 
							["DPS"] = L["DPS"], 
							["DPSC"] = L["DPSC"], 
							["DPSCD"] = L["DPSCD"], 
							["DPM"] = L["DPM"], 
							["MPS"] = L["MPS"], 
							["ManaCost"] = L["ManaCost"], 
							["TrueManaCost"] = L["TrueManaCost"],
							["CastTime"] = L["Cast Time"],
						},
						get = function() return settings["DisplayType"] end,
						set = function(info, v) 
								settings.FullText = (v == "CastTime")
								settings["DisplayType"] = v
								self:UpdateAB() 
						end,
					},
					CastsLeftDmg = {
						type = 'toggle',
						name = L["Damage spell casts"],
						desc = L["Actionbar displays the amount of damage spell casts left with current mana pool."],
						order = 5,
						hidden = playerMelee or playerClass == "DEATHKNIGHT",
						get = function() return settings["CastsLeftDmg"] end,
						set = function(info, v)
							settings.CastsLeftDmg = v
							if v then
								if not ManaBucket then
									ManaBucket = self:RegisterBucketEvent("UNIT_MANA", 2)
								end
							else
								if ManaBucket and not settings.CastsLeft then
									self:UnregisterBucket(ManaBucket)
									ManaBucket = nil
								end
							end
							self:UpdateAB()
						end,
					},
					CastsLeft = {
						type = 'toggle',
						name = L["Healing spell casts"],
						desc = L["Actionbar displays the amount of healing spell casts left with current mana pool."],
						order = 6,
						disabled = not playerHealer,
						hidden = playerMelee or playerClass == "DEATHKNIGHT",
						get = function() return settings["CastsLeft"] end,
						set = function(info, v)
							settings.CastsLeft = v
							if v then
								if not ManaBucket then
									ManaBucket = self:RegisterBucketEvent("UNIT_MANA", 2)
								end
							else
								if ManaBucket and not settings.CastsLeftDmg then
									self:UnregisterBucket(ManaBucket)
									ManaBucket = nil
								end
							end
							self:UpdateAB()				
						end,
					},					
					ABTitle = {
						type = 'header',
						name = L["Actionbar"],
						order = 10,
					},					
					UpdateAB = {
						type = 'execute',
						name = L["Update"],
						desc = L["Forces an update to the actionbar."],
						order = 11,
						func = function() self.visualChange = true; self:UpdateAB() end,
					},
					Update = {
						type = 'multiselect',
						name = L["Actionbar"],
						order = 22,
						values = {
								["ABText"] = L["Actionbar text on/off"],
								["DefaultAB"] =	L["Default actionbar support"],
								["SwapCalc"] = L["Swap calculation for double purpose abilities"],						
								["UpdateShift"] = L["Update the actionbar with shift key"],
								["UpdateAlt"] = L["Update the actionbar with alt key"],
								["UpdateCtrl"] = L["Update the actionbar with ctrl key"],
							},
						get = function(v, k) return settings[k] end,
						set = function(info, k, v)
							if (k == "UpdateShift") or (k == "UpdateAlt") or (k == "UpdateCtrl") then
								settings[k] = v
								if v and not ModStateEvent then
									self:RegisterEvent("MODIFIER_STATE_CHANGED")
									ModStateEvent = true
								elseif ModStateEvent and not settings.UpdateAlt and not settings.UpdateCtrl and not settings.UpdateShift then
									self:UnregisterEvent("MODIFIER_STATE_CHANGED")
									ModStateEvent = false
								end
							else
								self.visualChange = true
								if (k == "DefaultAB") then
									if v then
										settings.DefaultAB = true
										self:UpdateAB()
									else
										self:UpdateAB(nil, nil, true)
										settings.DefaultAB = false
									end								
								else
									settings[k] = v
									self:UpdateAB()
								end
							end

						end,
					},			
					FontSize = {
						type = 'range',
						name = L["Actionbar text size"],
						min = 6,
						max = 20,
						step = 1,
						order = 30,
						get =  function() return settings["FontSize"] end,
						set = function(i, v)
							settings["FontSize"] = v
							self.visualChange = true
							self:UpdateAB()
						end,
					},
					FontXPosition = {
						type = 'range',
						name = L["Actionbar text horizontal position"],
						min = -5,
						max = 5,
						step = 1,
						order = 35,
						get =  function() return settings["FontXPosition"] end,
						set = function(i, v)
							settings["FontXPosition"] = v
							self.visualChange = true
							self:UpdateAB()
						end,
					},
					FontYPosition = {
						type = 'range',
						name = L["Actionbar text vertical position"],
						min = -15,
						max = 30,
						step = 1,
						order = 40,
						get =  function() return settings["FontYPosition"] end,
						set = function(i, v)
							settings["FontYPosition"] = v
							self.visualChange = true
							self:UpdateAB()
						end,
					},	
					FontEffect = {
						type = 'select',
						name = L["Actionbar text effect"],
						values = { ["OUTLINE"] = L["Outline"], ["THICKOUTLINE"] = L["ThickOutline"], [""] = L["None"] },
						order = 65,
						get =	function() return settings["FontEffect"] end,
						set = 	function(i, v)
							settings["FontEffect"] = v
							self.visualChange = true
							self:UpdateAB()
						end, 
					},				
					FontColorDmg = {
						type = 'color',
						name = L["Damage text color"],
						desc = L["Set the actionbar damage text color"],
						order = 70,
						get = function() return settings.FontColorDmg.r, settings.FontColorDmg.g, settings.FontColorDmg.b end,
						set = function(info,rr,gg,bb) 
							settings.FontColorDmg.r = rr
							settings.FontColorDmg.g = gg
							settings.FontColorDmg.b = bb
							self.visualChange = true
							self:UpdateAB()
						end,
					},
					FontColorHeal = {
						type = 'color',
						name = L["Heal text color"],
						desc = L["Set the actionbar heal text color."],
						order =75,
						get = function() return settings.FontColorHeal.r, settings.FontColorHeal.g, settings.FontColorHeal.b end,
						set = function(info,rr,gg,bb) 
							settings.FontColorHeal.r = rr
							settings.FontColorHeal.g = gg
							settings.FontColorHeal.b = bb
							self.visualChange = true
							self:UpdateAB()
						end,
					},
					FontColorMana = {
						type = 'color',
						name = L["Mana text color"],
						desc = L["Set the actionbar mana text color."],
						order = 80,
						hidden = playerMelee or playerClass == "DEATHKNIGHT",
						get = function() return settings.FontColorMana.r, settings.FontColorMana.g, settings.FontColorMana.b end,
						set = function(info,rr,gg,bb) 
							settings.FontColorMana.r = rr
							settings.FontColorMana.g = gg
							settings.FontColorMana.b = bb
							self.visualChange = true
							self:UpdateAB()
						end,
					},				
				},
			},
			Tooltip = {
				type = 'group', 
				name = L["Tooltip"], 
				order = 55,
				args = {
					DefaultColor = {
						type = 'toggle',
						name = L["Default tooltip colors"],
						order = 5,
						get = function() return settings["DefaultColor"] end,
						set = function(i, v) settings["DefaultColor"] = v end,
					},
					Hints = {
						type = 'toggle',
						name = L["Display tooltip hints"],
						order = 10,
						get = function() return settings["Hints"] end,
						set = function(i, v) settings["Hints"] = v end,
					},			
					DisplayTooltip = {
						type = 'select',
						name = L["Display tooltip"] ,
						values = { 
							["Always"] = L["Always"],
							["Never"] = L["Never"],
							["Combat"] = L["Disable in combat"],
							["Alt"] = L["With Alt"],
							["Ctrl"] = L["With Ctrl"],
							["Shift"]= L["With Shift"]					
						},
						order = 15,
						get = function() return settings.Tooltip end,
						set = function(info, k, v)
							settings.Tooltip = k
						end,
					},
					Tooltip = {
						type = 'multiselect', 
						name = L["Display"],
						order = 20,
						values = {
							["Coeffs"] = L["Show coefficients"],
							["DispCrit"]= L["Show crit %"],
							["DispHit"] = L["Show hit %"],
							["AvgHit"] = L["Show avg and hit range"],
							["AvgCrit"] = L["Show avg crit and crit range"],
							["Ticks"] = L["Show per tick"],
							["Total"] = L["Show avg total"],
							["Extra"] = L["Show additional effects"],
							["Next"] = L["Show stat increase values"],
							["DPS"] = L["Show DPS/HPS"],
							["DPM"] = L["Show DPM/HPM"],
						},
						get = function(v, k) return settings[k] end,
						set = function(info, k, v) self:ClearTooltip(); settings[k] = v end,
					},
				},
					
			},
			Calculation = {
				type = 'group', 
				name = L["Calculation"], 
				desc = L["Calculation options."], 
				order = 60,
				args = {
					General = {
						type = 'multiselect',
						name = L["Calculation"],
						values = {
							["HitCalc"] = L["Hit calculation"],
						},
						order = 10,
						get = function(v, k) return settings[k] end,
						set = function(info, k, v)
							settings[k] = v
							self:UpdateAB()
						end,
					},					
					TargetAmount = {
						type = 'range',
						name = L["Amount of targets"],
						desc = L["Select the maximum amount of targets for your AoE abilities."],
						min = 1,
						max = 10,
						step = 1,
						order = 45,
						get = function() return settings["TargetAmount"] end,
						set = function(info, v) settings["TargetAmount"] = v; self:UpdateAB() end,
					},
					ComboPoints = {
						type = 'range',
						name = L["Combo points"],
						desc = L["Manually set the amount of calculated combo points. When 0 is selected, the calculation is based on the current amount."],
						min = 0,
						max = 5,
						step = 1,
						order = 50,
						hidden = (playerClass ~= "ROGUE" and playerClass ~= "DRUID"),
						get = function() return settings["ComboPoints"] end,
						set = function(info,v) settings["ComboPoints"] = v; self:UpdateAB() end,
					},																		
					MitigationTitle = {
						type = 'header',
						name = L["Target mitigation calculation"],
						order = 55,
					},
					HitTarget = {			
						type = 'toggle',
						name = L["Calculate by current target"],
						desc = L["Calculates by current target level. If nothing is targeted, the manually set level is used."],
						order = 60,
						get = function() return settings["HitTarget"] end,
						set = function(info, v) settings["HitTarget"] = v; self:UpdateAB() end,
					},										
					BossArmor = {
						type = 'toggle',
						name = L["Boss armor calculation"],
						desc = string.format(L["Automatically use %d armor for bosses"], 10643),
						order = 65,
						disabled = playerCaster,
						get = function() return settings["BossArmor"] end,
						set = function(info,v) settings["BossArmor"] = v; self:UpdateAB() end,
					},
					TargetLevel = {
						type = 'range',
						name = L["Target level"],
						desc = L["Target level compared to your level. Set as 3 for boss calculation."],
						min = -10,
						max = 10,
						step = 1,
						order = 70,
						get = function() return settings["TargetLevel"] end,
						set = function(info, v) settings["TargetLevel"] = v; self:UpdateAB() end,
					},					
					Armor = {
						type = 'range',
						name = L["Target armor"],
						desc = L["Estimated target armor for non-boss enemies. Compare to tooltip numbers for desired mitigation."],
						min = 0,
						max = 50000,
						step = 1,
						order = 75,
						disabled = playerCaster,
						get = function() return settings["Armor"] end,
						set = function(info,v) settings["Armor"] = v; self:UpdateAB() end,
					},
					Resilience = {
						type = 'range',
						name = L["Target resilience"],
						desc = L["Input your target's resilience."],
						min = 0,
						max = 1500,
						step = 1,
						order = 80,
						get = function() return settings["Resilience"] end,
						set = function(info,v) settings["Resilience"] = v; self:UpdateAB() end,
					},					
				},
			},
			Custom = {
				type = 'group', 
				name = L["Modify Stats"], 
				--desc = L["Calculation options."], 
				order = 95,
				args = {
					Custom = {
						type = 'toggle',
						name = L["Use custom stats"],
						desc = L["Manually set stats are used in calculations."],
						order = 50,
						get = function() return settings["Custom"] end,
						set = function(info, v) settings["Custom"] = v; self:UpdateAB() end,
					},
					CustomAdd = {
						type = 'toggle',
						name = L["Add custom stats"],
						desc = L["Manually set stats are added to your stats in calculations."],
						order = 55,
						get = function() return settings["CustomAdd"] end,
						set = function(info, v) settings["CustomAdd"] = v; self:UpdateAB() end,
					},
					SpellPower = {
						type = 'range',
						name = L["Spell power"],
						desc = L["Input spell power to use in calculations."],
						min = -20000,
						max = 20000,
						step = 1,
						order = 105,
						hidden = playerMelee or (playerClass == "DEATHKNIGHT"),
						get = function() return settings["SpellDamage"] end,
						set = DrD_Set("SpellDamage"),
					},
					AP = {
						type = 'range',
						name = L["AP"] .. "/" .. L["RAP"],
						desc = L["Input attack power to use in calculations."],
						min = -20000,
						max = 20000,
						step = 1,
						order = 106,
						hidden = playerCaster,
						get = function() return settings["AP"] end,
						set = DrD_Set("AP"),
					},					
					CritRating = {
						type = 'range',
						name = L["Critical strike rating"],
						desc = L["Input critical strike rating to use in calculations."],
						min = -2000,
						max = 2000,
						step = 1,
						order = 110,
						get = function() return settings["CritRating"] end,
						set = DrD_Set("CritRating"),					
					},
					HitRating = {
						type = 'range',
						name = L["Hit rating"],
						desc = L["Input hit rating to use in calculations."],
						min = -1000,
						max = 1000,
						step = 1,
						order = 115,
						get = function() return settings["HitRating"] end,
						set = DrD_Set("HitRating"),	
					},
					HasteRating = {
						type = 'range',
						name = L["Spell haste rating"],
						desc = L["Input haste rating to use in calculations."],
						min = -5000,
						max = 5000,
						step = 1,
						order = 120,
						hidden = playerMelee or (playerClass == "DEATHKNIGHT"),
						get = function() return settings["HasteRating"] end,
						set = DrD_Set("HasteRating"),	
					},
					ManaPer5 = {
						type = 'range',
						name = L["Mana per 5"],
						min = -5000,
						max = 5000,
						step = 1,
						desc = L["Input MP5 to use in calculations."],
						order = 125,
						hidden = playerMelee or (playerClass == "DEATHKNIGHT"),
						get = function() return settings["MP5"] end,
						set = DrD_Set("MP5"),	
					},
				},
			},
			Aura = {
				type = 'group',
				name = L["Modify Buffs/Debuffs"],
				desc = L["Choose what buffs/debuffs to always include into calculations."],
				order = 100,
				args = {
					Player = {
						type = 'multiselect',
						name = L["Player"],
						desc = L["Choose player buffs/debuffs."],
						order = 1,
						values = {},
						get = function(v, k) return settings["PlayerAura"][k] end,
						set = function(info, k, v)
							local cat = self.PlayerAura[k].Category
							if v and cat then
								for n in pairs( settings.PlayerAura ) do
									local ocat = self.PlayerAura[n].Category
									if ocat and (cat == ocat) then
										settings.PlayerAura[n] = nil
									end
								end						
							end					
							settings["PlayerAura"][k] = v or nil
							self:UpdateAB() 
						end,		
					},				
					Target = {
						type = 'multiselect',
						name = L["Target"],
						desc = L["Choose target buffs/debuffs."],
						values = {},
						order = 3,
						get = function(v, k) return settings["TargetAura"][k] end,
						set = function(info, k, v)
							local cat = self.TargetAura[k].Category
							if v and cat then
								for n in pairs( settings.TargetAura ) do
									local ocat = self.TargetAura[n].Category
									if ocat and (cat == ocat) then
										settings.TargetAura[n] = nil
									end
								end						
							end					
							settings["TargetAura"][k] = v or nil
							self:UpdateAB() 
						end,						
					},
					Consumables = {
						type = 'multiselect',
						name = L["Consumables"],
						desc = L["Choose consumables to include."],
						values = {},
						order = 2,
						get = function(v, k) return settings["Consumables"][k] end,
						set = function(info, k, v)
							local cat = self.Consumables[k].Category
							local cat2 = self.Consumables[k].Category2
							if v and cat then
								for n in pairs( settings.Consumables ) do
									local ocat = self.Consumables[n].Category
									local ocat2 = self.Consumables[n].Category2
									if ocat and (cat == ocat or cat2 and cat2 == ocat or ocat2 and (cat == ocat2 or cat2 and ocat2 == cat2)) then
										settings.Consumables[n] = nil
									end
								end
							end
							settings["Consumables"][k] = v or nil 
							self:UpdateAB() 
						end,								
					},				
				},
			},
			Talents = { 
				type = "group", 
				name = L["Modify Talents"], 
				desc = L["Modify talents manually. Modified talents are not saved between sessions."], 
				order = 150, 
				args = {
					Reset = { 
						type = "execute", 
						name = L["Reset Talents"],
						desc = L["Reset talents to your current talent configuration."],
						order = 0,
						disabled = self.CustomTalents,
						func = function() self:UpdateTalents() end,
					},
					Remove = { 
						type = "execute", 
						name = L["Remove Talents"],
						desc = L["Removes all talents from your current configuration."],
						order = 1,
						disabled = self.CustomTalents,
						func = function() 
							DrD_ClearTable(talents)
							self:UpdateTalents(true) 
						end,
					},					
				} ,
			},			
		},
	}
	local optionsTable = self.options.args.General.args	
	if playerCaster or playerHybrid then
		local tree = optionsTable.Tooltip.args.Tooltip.values
		tree["PlusDmg"] = L["Show efficient spellpower"]
		tree["CompareDmg"] = L["Compare spell power"]
		tree["CompareCrit"] = L["Compare critical strike rating"]
		tree["CompareHit"] = L["Compare hit rating"]
		tree["CompareHaste"] = L["Compare haste rating"]
		tree["Doom"] = L["Show damage/healing until OOM"]
		tree["Casts"] = L["Show total casts and time until OOM"]
		tree["ManaUsage"] = L["Show additional mana usage information"]
		tree = optionsTable.Calculation.args.General.values
		tree["TwoRoll"] = L["Two roll calculation"] .. " (" .. L["Spells"] .. ")"
		if not (playerClass == "DEATHKNIGHT") then
			tree["ManaConsumables"] = L["Include mana consumables"]
		end
	end
	if playerMelee or playerHybrid then
		local tree = optionsTable.Tooltip.args.Tooltip.values
		tree["DPP"] = L["Show damage per power"]
		tree = optionsTable.Calculation.args.General.values
		tree["CritDepression"] = L["Crit depression calculation"]
		tree["Dodge"] = L["Dodge calculation"]
		tree["Parry"] = L["Parry calculation"]
		tree["Glancing"] = L["Glancing blow calculation"]				
		tree["TwoRoll_M"] = L["Two roll calculation"] .. " (" .. L["Melee"] .. "/" .. L["Ranged"] .. ")"	
	end
	if playerClass == "DEATHKNIGHT" then
		local types = optionsTable.Actionbar.args.DisplayType.values
		types["PowerCost"] = nil
		types["DPM"] = nil
		types = optionsTable.Actionbar.args.DisplayType2.values
		types["DPM"] = nil
		types["MPS"] = nil
		types["ManaCost"] = nil
		types["TrueManaCost"] = nil
		types["CastTime"] = nil
	end
	if LSM then
		optionsTable.Actionbar.args.Font = {
			type = 'select',
			name = L["Actionbar text font"],
			desc = L["Actionbar text font"],
			values = LSM:HashTable("font"),
			order = 60,
			get =  function()
				if LSM:GetGlobal("font") then
					return LSM:GetGlobal("font")
				end
				for k, v in pairs(LSM:HashTable("font")) do 
					if settings["Font"] == v then 
						return k 
					end 
				end
			end,
			set = function(info, v, k)
				if not LSM:GetGlobal("font") then
					settings.Font = LSM:Fetch("font",v)
					DrD_Font = settings.Font
					self.visualChange = true
					self:UpdateAB()
				end
			end,
		}		
		if AceGUISharedMediaWidgets then
			optionsTable.Actionbar.args.Font.dialogControl = 'LSM30_Font'
		end
		if LSM:GetGlobal("font") then
			DrD_Font = LSM:Fetch("font")
		end
		
	end
	local auraTable = optionsTable.Aura.args.Player.values
	local double = {}
	for k,v in pairs( self.PlayerAura ) do
		if not v.Update and not v.NoManual then
			local manual = v.Manual
			if not manual or not double[manual] then
				if v.ID then
					auraTable[k] = "|T" .. select(3,GetSpellInfo(v.ID)) .. ":17:17:-2:0|t" .. (manual or k)			
				--elseif GetSpellInfo((manual or k)) then
				--	auraTable[k] = "|T" .. select(3,GetSpellInfo((manual or k))) .. ":17:17:-2:0|t" .. (manual or k)			
				else	
					auraTable[k] = manual or k
				end
				double[(manual or k)] = true
			else
				settings["PlayerAura"][k] = nil
			end
		else
			settings["PlayerAura"][k] = nil
		end
	end
	auraTable = optionsTable.Aura.args.Target.values
	double = {}
	for k,v in pairs( self.TargetAura ) do
		if not v.Update and not v.NoManual then
			local manual = v.Manual
			if not manual or not double[manual] then
				if v.ID then
					auraTable[k] = "|T" .. select(3,GetSpellInfo(v.ID)) .. ":17:17:-2:0|t" .. (manual or k)
				--elseif GetSpellInfo((manual or k)) then
				--	auraTable[k] = "|T" .. select(3,GetSpellInfo((manual or k))) .. ":17:17:-2:0|t" .. (manual or k)
				else
					auraTable[k] = manual or k
				end
				double[(manual or k)] = true
			else
				settings["TargetAura"][k] = nil
			end
		else
			settings["TargetAura"][k] = nil
		end
	end	
	auraTable = optionsTable.Aura.args.Consumables.values
	for k,v in pairs( self.Consumables ) do
		if v.Category == "Food" then
			auraTable[k] = "|TInterface\\Icons\\Spell_Misc_Food:17:17:-2:0|t" .. k
		elseif select(10,GetItemInfo(k)) then
			auraTable[k] = "|T" .. select(10,GetItemInfo(k)) .. ":17:17:-2:0|t" .. k
		else
			auraTable[k] = k
		end
	end
	local talentTable = optionsTable.Talents.args
	for t = 1, GetNumTalentTabs() do	
		for i = 1, GetNumTalents(t) do
			local talentName, icon, _, _, _, maxRank = GetTalentInfo(t, i)
			if talentInfo[talentName] and not talentInfo[talentName].NoManual then
				talentTable[(string_gsub(talentName," +", ""))] = {
					type = 'range',
					name = "|T" .. icon .. ":20:20:-5:0|t" .. talentName,
					disabled = self.CustomTalents,
					min = 0,
					max = maxRank,
					step = 1,
					order = 3 + i + (t-1) * 50,
					get = 	function() return talents[talentName] or 0 end,
					set = 	function(info, v)
							local value = math_floor(v+0.5)
							if value == 0 then value = nil end
							talents[talentName] = value
							self:UpdateTalents(true)
						end,
				}
			end
		end
		talentTable[("Tab" .. t)] = {
			type = 'header',
			name = --[["|T" .. select(2,GetTalentTabInfo(t)) .. ":30:30:-7:0|t" ..--]] GetTalentTabInfo(t),
			order = 2 + (t-1) * 49,
		}	
	end		
end

--EVENTS:
--Event for weapon buff (poisons, spellstones etc) updates
local mbuff, obuff
function DrDamage:UNIT_INVENTORY_CHANGED(units)
	if units["player"] then
		local buff = self:GetWeaponBuff()
		local buff2 = self:GetWeaponBuff(true)
		if buff ~= mbuff or buff2 ~= obuff then
			mbuff = buff
			obuff = buff2
			self:CancelTimer(self.fullUpdate, true)
			self:UpdateAB()
		end
	end
end

--REQUIRES DELAY:

function DrDamage:ACTIONBAR_PAGE_CHANGED()
	self.visualChange = true
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.2)
	end
end

--Triggers after an ability is placed on the actionbar
function DrDamage:ACTIONBAR_HIDEGRID()
	self.visualChange = true
	if not self.fullUpdate and (GetCursorInfo() == "spell" or GetCursorInfo() == "macro") then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.2)
	end
end

--Shapeshift and update for certain auras (like Shadowform)
function DrDamage:UPDATE_SHAPESHIFT_FORM()
	self.visualChange = true
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 1)
	end
end

--Castsequence macro update triggers
local MacroTimer, Trigger
local shaman = (playerClass == "SHAMAN")
function DrDamage:MacroTrigger()
	Trigger = nil
	if not shaman then
		self:UnregisterEvent("ACTIONBAR_SLOT_CHANGED")
	end
end
function DrDamage:EXECUTE_CHAT_LINE()
	self:CancelTimer(MacroTimer, true)
	Trigger = true
	MacroTimer = self:ScheduleTimer("MacroTrigger", 0.5)
	if not shaman then
		self:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
	end
end
function DrDamage:ACTIONBAR_SLOT_CHANGED(_, id)
	if Trigger then
		local gtype, pid = GetActionInfo(id)
		if gtype == "macro" then
			local name = GetMacroSpell(pid)
			if name and spellInfo[name] then
				self.visualChange = true
				if not self.fullUpdate then
					self:ScheduleTimer("UpdateAB", 0.5, nil, id)
					Trigger = nil
				end
			end
		end
	elseif shaman and (id == 133 or id == 134 or id == 135 or id == 136) then
		self.visualChange = true
		if not self.fullUpdate then
			self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.25, nil, id)
		end	
	end
end

--DELAYED EVENTS FOR INCREASED PERFORMANCE

--Main-hand, Off-hand, Ranged slot and meta gem updates
DrDamage.Damage_critMBonus = 0
DrDamage.Healing_critMBonus = 0
local MetaGemTimer
function DrDamage:PLAYER_EQUIPMENT_CHANGED(event, slot)
	updateSetItems = true
	if self.Melee_InventoryChanged then
		self:Melee_InventoryChanged((slot == 16), (slot == 17), (slot == 18))
	end
	if not MetaGemTimer then
		MetaGemTimer = self:ScheduleTimer("MetaGems", 0.20)
	end	
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.3)
	end
end
function DrDamage:MetaGems()
	MetaGemTimer = nil
	self.Damage_critMBonus = 0
	self.Healing_critMBonus = 0
	local helm = GetInventoryItemLink("player", 1)
	if helm then
		for i = 1, 3 do 
			local mgem = tonumber(string_match(select(2, GetItemGem(helm, i)) or "",".-item:(%d+).*"))
			if mgem == 34220 or mgem == 41285 or mgem == 32409 or mgem == 41398 then
				if not self:IsMetaGemInactive() then
					self.Damage_critMBonus = 0.03
				end
				break
			elseif mgem == 41376 then
				if not self:IsMetaGemInactive() then
					self.Healing_critMBonus = 0.03
				end
				break
			end
		end
	end
end

--Event for checking if the meta gem is greyed out in the head slot tooltip
function DrDamage:IsMetaGemInactive()
	if GT:SetInventoryItem("player", 1) then
		for j = 1, GT:NumLines() do 
			if GT:GetLine(j) and string_find(GT:GetLine(j), "|cff808080.*808080") then
				return true
			end
		end
	end
end

--Event for skills and talents that update on certain health percentages
local pHealth, tHealth = 1, 1
function DrDamage:UNIT_HEALTH(units)
	if PlayerHealth and units["player"] then
		local value = PlayerHealth[1]
		local health = UnitHealth("player")/UnitHealthMax("player")
		if (health < value) ~= (pHealth < value) then
			self:UpdateAB(PlayerHealth[value])
		end
		pHealth = health
	end
	if units["target"] then
		local health = UnitHealth("target")/UnitHealthMax("target")
		for i=1,#TargetHealth do
			local value = tonumber(TargetHealth[i]) or UnitHealth("player")/UnitHealthMax("player")
			if (health < value) ~= (tHealth < value) then
				if TargetHealth[value] then 
					self:UpdateAB(TargetHealth[value])
				else
					self:UpdateAB()
					tHealth = health
					return
				end
			end
		end
		tHealth = health
	end
end

function DrDamage:PLAYER_TARGET_CHANGED()
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.5)
	end
	self:TargetAuraUpdate(true)
	tHealth = TargetHealth and (UnitHealth("target")/UnitHealthMax("target"))
end

--Triggers after macros are updated
function DrDamage:UPDATE_MACROS()
	self.visualChange = true
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 0.75)
	end
end

--Triggers when new ranks of spells are learned
function DrDamage:LEARNED_SPELL_IN_TAB()
	if not self.fullUpdate then
		self.fullUpdate = self:ScheduleTimer("UpdateAB", 1.5)
	end
end

local PlayerAuraTimer, TargetAuraTimer
function DrDamage:UNIT_AURA(event, unit)
	if unit == "player" then
		if not PlayerAuraTimer then
			PlayerAuraTimer = self:ScheduleTimer("PlayerAuraUpdate", 0.75)
		end
	elseif unit == "target" then
		if not TargetAuraTimer then
			TargetAuraTimer = self:ScheduleTimer("TargetAuraUpdate", 1.5)
		end	
	end
end

local PlayerAura = DrDamage.PlayerAura
local TargetAura = DrDamage.TargetAura
local oPlayerAura, nPlayerAura = {}, {}
local oTargetAura, nTargetAura = {}, {}

function DrDamage:PlayerAuraUpdate()
	PlayerAuraTimer = nil
	if self.fullUpdate then
		return
	end	
	if dmgMod ~= select(7, UnitDamage("player")) then
		self:UpdateAB()
		return
	end
	if self.Caster_CheckBaseStats and self:Caster_CheckBaseStats() then
		self:UpdateAB()
		return
	end
	if self.Melee_CheckBaseStats and self:Melee_CheckBaseStats() then
		self:UpdateAB()
		return
	end
	--[[ Enable if ever needed
	if self.Calculation["PlayerAura"] then
		local update, spell = self.Calculation["PlayerAura"]
		if update then
			return self:UpdateAB(spell)
		end
	end
	--]]
	for i=1,40 do
		local name, rank, texture, count = UnitBuff("player",i)
		if name then
			if PlayerAura[name] then			
				nPlayerAura[name] = rank .. count
			end
		else
			break
		end
	end
	for i=1,40 do
		local name, rank, texture, count = UnitDebuff("player",i)
		if name then
			if PlayerAura[name] then
				nPlayerAura[name] = rank .. count
			end
		else 
			break
		end
	end
	local buffName, multi
	--Buff/debuff gained or count/rank changed
	for k,v in pairs(nPlayerAura) do
		if not oPlayerAura[k] or oPlayerAura[k] ~= v then
			multi = buffName
			buffName = k
		end
	end	
	--Buff/debuff lost
	for k,v in pairs(oPlayerAura) do
		if not nPlayerAura[k] then
			multi = buffName
			buffName = k
		end
		oPlayerAura[k] = nil
	end
	--Copy new table to old and clear
	for k,v in pairs(nPlayerAura) do
		oPlayerAura[k] = v
		nPlayerAura[k] = nil
	end
	if buffName then
		self:UpdateAB(not multi and PlayerAura[buffName].Spells)
	end
end

function DrDamage:TargetAuraUpdate( changed )
	TargetAuraTimer = nil
	if self.fullUpdate and not changed then
		return
	end
	if self.Calculation["TargetAura"] then
		local update, spell = self.Calculation["TargetAura"]
		if update and not changed then
			return self:UpdateAB(spell)
		end
	end		
	if playerHealer then
		for i=1,40 do
			local name, rank, texture, count, _, _, _, unit = UnitBuff("target",i)
			if name then
				if TargetAura[name] then
					if unit and TargetAura[name].SelfCastBuff then
						if UnitIsUnit("player",unit) then
							nTargetAura[name] = rank .. count
						end
					else
						nTargetAura[name] = rank .. count
						--nTargetAura[name.."|"..texture] = rank .. count
					end
				end
			else break end
		end
	end
	for i=1,40 do
		local name, rank, texture, count, _, _, _, unit = UnitDebuff("target",i)
		if name then
			if TargetAura[name] then
				if unit and TargetAura[name].SelfCast then
					if UnitIsUnit("player",unit) then
						nTargetAura[name] = rank .. count
					end
				else
					nTargetAura[name] = rank .. count
					--nTargetAura[name.."|"..texture] = rank .. count
				end
			end
		else break end
	end
	local buffName, multi
	--Buff/debuff gained or count/rank changed
	for k,v in pairs(nTargetAura) do
		if not oTargetAura[k] or oTargetAura[k] ~= v then
			multi = buffName
			buffName = k
			--buffName = string_match(k,"[^|]+")
		end
	end	
	--Buff/debuff lost
	for k,v in pairs(oTargetAura) do
		if not nTargetAura[k] then
			multi = buffName
			buffName = k
			--buffName = string_match(k,"[^|]+")
		end
		oTargetAura[k] = nil
	end
	--Copy new table to old and clear
	for k,v in pairs(nTargetAura) do
		oTargetAura[k] = v
		nTargetAura[k] = nil
	end		
	if TargetAura[buffName] and not changed then
		self:UpdateAB(not multi and TargetAura[buffName].Spells)
	end
end

--Event for actionbar modifier key updates
local lastState = GetTime()
local ModTimer
function DrDamage:MODIFIER_STATE_CHANGED(event, state)
	if (state == "LALT" or state == "RALT") and settings.UpdateAlt or (state == "LCTRL" or state == "RCTRL") and settings.UpdateCtrl or (state == "LSHIFT" or state == "RSHIFT") and settings.UpdateShift then
		if GetTime() - lastState < 0.3 then
			self:CancelTimer(ModTimer, true)
			ModTimer = false
			self.visualChange = false
		else
			ModTimer = self:ScheduleTimer("UpdateAB", 0.3)
			self.visualChange = true
		end
		lastState = GetTime()
	end
end

--Talent updates
local TalentTimer, TalentIterate
function DrDamage:UPDATE_TALENTS()
	self.visualChange = true
	if not TalentTimer then
		TalentTimer = self:ScheduleTimer("UpdateTalents", 1)
	end
end
function DrDamage:UpdateTalents(manual, all)
	TalentTimer = nil
	if not manual and not self.CustomTalents then
		DrD_ClearTable( talents )
		for t = 1, GetNumTalentTabs() do	
			for i = 1, GetNumTalents(t) do
				local talentName, _, _, _, currRank, maxRank = GetTalentInfo(t, i)
				if (currRank ~= 0 or all) and talentInfo[talentName] then
					talents[talentName] = all and maxRank or currRank
				end	    	    
			end
		end
	end
	--Load talents into appropriate spell data
	for _, spell in pairs( spellInfo ) do
		TalentIterate( spell["Name"], spell[0], spell["Talents"] )
		if spell["Secondary"] then
			TalentIterate( spell["Secondary"]["Name"], spell["Secondary"][0], spell["Secondary"]["Talents"] )
		end
	end
	if not self.fullUpdate then
		self:UpdateAB()
	end
end
TalentIterate = function( name, baseSpell, endtable )
	if not baseSpell or type(baseSpell) == "function" then return end
	DrD_ClearTable( endtable )
	local melee = baseSpell.Melee
	local school = baseSpell.School
	local s1,s2,s3
	if type(school) == "table" then
		s1 = school[1]
		s2 = school[2]
		s3 = school[3]
	else
		s1 = school or "Physical"
	end
	for talentName, talentRank in pairs( talents ) do
		local talentTable = talentInfo[talentName]
		for i=1,#talentTable do
			local talent = talentTable[i]
			if (melee and talent.Melee or talent.Caster and not melee or (not talent.Melee and not talent.Caster)) and not (talent.Specials and (baseSpell.AutoAttack or baseSpell.AutoShot)) then
				local spells = talent.Spells
				if DrD_MatchData(spells,name,s3) or not baseSpell.NoSchoolTalents and DrD_MatchData(spells, "All", s1) or not baseSpell.NoTypeTalents and DrD_MatchData(spells, s2) then
					local number = #endtable + 1
					local modtype = talent.ModType
					local multiply = talent.Multiply
					endtable["Multiply"..number] = multiply
					endtable[number] = type(talent.Effect) == "table" and talent.Effect[talentRank] or talent.Effect * talentRank
					if not modtype then
						endtable[("ModType"..number)] = multiply and "dmgM" or "dmgM_Add"
					elseif modtype == "SpellDamage" then
						endtable[("ModType"..number)] = multiply and "spellDmgM" or "spellDmgM_Add"
					else
						endtable[("ModType"..number)] = modtype
					end
					if talent.Mod then
						local mod = talent.Mod
						if DrD_MatchData(mod[1], name, "All", s1, s2, s3) then
							mod[6] = talentRank
							endtable["Mod"..number] = mod
						end
					end
				end
			end
		end
	end
end

local debug_flag
local SetItems = {}
function DrDamage:GetSetAmount(set)
	if debug_flag then
		return 8
	end
	if not updateSetItems and SetItems[set] then
		return SetItems[set]
	end
	
	if updateSetItems then
		DrD_ClearTable( SetItems )
		updateSetItems = false
	end
	
	local amount = 0
	local setData = self.SetBonuses[set]
	if setData then
		for _, itemID in ipairs(setData) do
			if IsEquippedItem(itemID) then
				amount = amount + 1
			end
		end
	end
	SetItems[set] = amount
	return amount
end

local Glyphs = {}
function DrDamage:HasGlyph(glyph)
	return debug_flag or Glyphs[glyph]
end

function DrDamage:UpdateGlyphs()
	DrD_ClearTable(Glyphs)
	for i=1,6 do
		local _, _, id = GetGlyphSocketInfo(i)
		if id then Glyphs[id] = true end
	end
end

local lines, saved = {}
local autoattack = GetSpellInfo(6603)

function DrDamage:ClearTooltip()
	if saved then
		for i = 1, #lines do
			lines[i]["Left"] = nil
			lines[i]["Right"] = nil
			lines[i]["Rl"] = nil
			lines[i]["Gl"] = nil
			lines[i]["Bl"] = nil			
			lines[i]["Rr"] = nil
			lines[i]["Gr"] = nil
			lines[i]["Br"] = nil
		end
		saved = nil
	end
end

function DrDamage:RedoTooltip(frame, name)
	if saved and name == saved then
		frame:AddLine(" ")
		for i=1, #lines do
			if lines[i]["Right"] then
				frame:AddDoubleLine(lines[i]["Left"], lines[i]["Right"], lines[i]["Rl"], lines[i]["Gl"], lines[i]["Bl"], lines[i]["Rr"], lines[i]["Gr"], lines[i]["Br"])
			elseif lines[i]["Left"] and lines[i]["Left"] ~= " " then
				frame:AddLine(lines[i]["Left"], lines[i]["Rl"], lines[i]["Gl"], lines[i]["Bl"])
			end
		end
		frame:Show()
		return true
	end
end

function DrDamage:SaveTooltip(name)
	if name == autoattack then
		saved = name
		for i=4, GameTooltip:NumLines() do
			local l = i-3
			if not lines[l] then
				lines[l] = {}
			end
			lines[l]["Left"] = _G[("GameTooltipTextLeft" .. i)]:GetText()
			lines[l]["Right"] = _G[("GameTooltipTextRight" .. i)]:GetText()
			lines[l]["Rl"], lines[l]["Gl"], lines[l]["Bl"] = _G[("GameTooltipTextLeft" .. i)]:GetTextColor()		
			lines[l]["Rr"], lines[l]["Gr"], lines[l]["Br"] = _G[("GameTooltipTextRight" .. i)]:GetTextColor()
		end
	end
end

function DrDamage:SetSpell( frame, slot )
	if settings.Tooltip == "Never" then return end
	if settings.Tooltip == "Combat" and InCombatLockdown() then return end
	if settings.Tooltip == "Alt" and not IsAltKeyDown() then return end
	if settings.Tooltip == "Ctrl" and not IsControlKeyDown() then return end
	if settings.Tooltip == "Shift" and not IsShiftKeyDown() then return end
	
	local name, rank = GetSpellName(slot,BOOKTYPE_SPELL)
	
	if self:RedoTooltip(frame, name) then return end
	
	if spellInfo[name] then
		local baseSpell = spellInfo[name][0]
		if type(baseSpell) == "function" then baseSpell = baseSpell() end

		if baseSpell and not baseSpell.NoTooltip then	
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterTooltip( frame, name, rank )
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeTooltip( frame, name, rank )
			end
		end
		self:SaveTooltip(name)
	end
end

function DrDamage:SetAction( frame, slot )
	if settings.Tooltip == "Never" then return end
	if settings.Tooltip == "Combat" and InCombatLockdown() then return end
	if settings.Tooltip == "Alt" and not IsAltKeyDown() then return end
	if settings.Tooltip == "Ctrl" and not IsControlKeyDown() then return end
	if settings.Tooltip == "Shift" and not IsShiftKeyDown() then return end
	
	local gtype, pid = GetActionInfo(slot)
	local name, rank
		
	if (gtype == "spell" and pid ~= 0) then
		name, rank = GetSpellName(pid, gtype)
	elseif gtype == "macro" then
		name, rank = GetMacroSpell(pid)
	end
	
	--if self:RedoTooltip(name) then return end
	
	if name and spellInfo[name] then
		local baseSpell = spellInfo[name][0]
		if type(baseSpell) == "function" then baseSpell = baseSpell() end

		if baseSpell and not baseSpell.NoTooltip then
			rank = rank or select(2,GetSpellInfo(name))
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterTooltip( frame, name, rank )
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeTooltip( frame, name, rank )
			end
		end
		--self:SaveTooltip(name)
	end
end

function DrDamage:SetShapeshift( frame, slot )
	if settings.Tooltip == "Never" then return end
	if settings.Tooltip == "Combat" and InCombatLockdown() then return end
	if settings.Tooltip == "Alt" and not IsAltKeyDown() then return end
	if settings.Tooltip == "Ctrl" and not IsControlKeyDown() then return end
	if settings.Tooltip == "Shift" and not IsShiftKeyDown() then return end
	
	local name, rank = frame:GetSpell()
	
	if name and spellInfo[name] then
		local baseSpell = spellInfo[name][0]
		if type(baseSpell) == "function" then baseSpell = baseSpell() end

		if baseSpell and not baseSpell.NoTooltip then	
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterTooltip( frame, name, rank )
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeTooltip( frame, name, rank )
			end
		end
	end	
end

local DrD_ProcessButton, DrD_CreateText
function DrDamage:UpdateAB(spell, uid, disable)
	--	Used for debugging updates.
	--[[
	if spell then
		self:Print( "Update AB: Spell" )
		
		if type( spell ) == "table" then
			for k in pairs( spell ) do
				self:Print( "Updating: " .. k )
			end
		else
			self:Print( "Updating: " .. spell )
		end
	elseif uid then
		self:Print( "Update AB: ID" )
		
	else
		self:Print( "Update AB: All" )
	end
	--]]
	
	--Enable if needed
	--if ABfunc then
	--	ABfunc()
	--end
	if ABtable then
		for name, func in pairs(ABtable) do
			DrD_ProcessButton(_G[name], func, spell, uid)
		end	
	end
	if not ABdisable and settings.DefaultAB then
		for i, name in ipairs(ABobjects) do
			DrD_ProcessButton(name, ActionButton_GetPagedID, spell, uid, disable)
		end
	end				
	self.fullUpdate = nil
	self.visualChange = nil
	dmgMod = select(7, UnitDamage("player"))
	self:ClearTooltip()
end

DrD_ProcessButton = function(button, func, spell, uid, disable)
	if not button then return end
	if not settings.ABText or disable then
		if button.drd then
			button.drd:Hide() 
		end		
		return 
	end
	if button:IsVisible() then
		if not spell and not uid then
			local frame = button.drd
			if frame then frame:Hide() end
		end	
		local id, name, rank = func(button)
		if id and (not HasAction(id) or uid and uid ~= id) then return end
		DrDamage:CheckAction(button, spell, id, name, rank)
	end
end

function DrDamage:CheckAction(button, spell, id, name, rank)
	local gtype, pid
	if id then
		gtype, pid = GetActionInfo(id)

		if gtype == "spell" and (pid > 0) then
			name, rank = GetSpellName(pid, gtype)
		elseif gtype == "macro" then
			name, rank = GetMacroSpell(pid)
		end
	end
	if name then
		if spell and not DrD_MatchData(spell, name) then
			return
		end
		if self.ClassSpecials[name] then
			rank = rank and tonumber(string_match(rank,"%d+"))
			return DrD_CreateText(button, self.ClassSpecials[name](rank))
		end
		if spellInfo[name] then
			local baseSpell = settings.SwapCalc and spellInfo[name]["Secondary"] and spellInfo[name]["Secondary"][0] or spellInfo[name][0]
			if type(baseSpell) == "function" then baseSpell = baseSpell() end

			if baseSpell then
				if gtype == "macro" and button then
					local macroText = _G[button:GetName().."Name"]
					if macroText then macroText:Hide() end
				end			

				local text, healingSpell

				if playerCaster or playerHybrid and not baseSpell.Melee then
					healingSpell = DrD_MatchData(baseSpell.School, "Healing")

					if (settings.CastsLeft and healingSpell or settings.CastsLeftDmg and not healingSpell) and not baseSpell.NoDPM and playerClass ~= "DEATHKNIGHT" then
						local _, _, _, manaCost = GetSpellInfo(name,rank)
						text = manaCost and (manaCost > 0) and math_floor(UnitPower("player",0) / manaCost) or "\226\136\158"
					else
						text = self:CasterCalc(name, rank)								
					end
				elseif playerMelee or playerHybrid and baseSpell.Melee then
					text = self:MeleeCalc(name, rank)
				end
				
				return DrD_CreateText(button, text, healingSpell)
			end
		end
	end
end

DrD_CreateText = function(button, text, healing, mana, full, r, g, b )
	if not text then 
		return 
	end
	if not full and not settings.FullText and type(text) == "number" then 
		text = math_floor(text + 0.5) 
	end
	if not r and (DrDamage.visualChange or not button or not button.drd) then
		local color = healing and settings.FontColorHeal or mana and settings.FontColorMana or settings.FontColorDmg
		r,g,b = color.r, color.g, color.b
	end
	if button then
		local drd = button.drd
		if drd then
			if DrDamage.visualChange then
				drd:SetFont(DrD_Font, settings.FontSize, settings.FontEffect)
				drd:SetPoint("BOTTOMLEFT", button, "BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
				drd:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
				drd:SetTextColor(r,g,b)	
			end
			drd:SetText(text)
			drd:Show()
		else
			drd = button:CreateFontString("drd", "OVERLAY")
			button.drd = drd
			drd:SetPoint("BOTTOMLEFT", button, "BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
			drd:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
			drd:SetFont(settings.Font, settings.FontSize, settings.FontEffect)
			drd:SetJustifyH("CENTER")
			drd:SetTextColor(r,g,b)
			drd:SetText(text)
			drd:Show()
		end
	else
		return "|cff".. string_format("%02x%02x%02x", r * 255, g * 255, b * 255) .. text .. "|r"
	end	
end


function DrDamage.BuffCalc( data, calculation, ActiveAuras, Talents, baseSpell, buffName, index, apps, texture, rank, target )
	local school = data.School
	if DrD_MatchData(data.Spells, calculation.name)
	or not calculation.healingSpell and (not school and not data.Spells or DrD_MatchData(school, calculation.school))
	or calculation.healingSpell and DrD_MatchData(school, "Healing")
	or (DrD_MatchData(school, "Spells") or DrD_MatchData(school, "Damage Spells") and not calculation.healingSpell) and (calculation.group ~= "Physical") and not calculation.physical
	or DrD_MatchData(school, calculation.group, calculation.subType) then
		if data.Not and DrD_MatchData(data.Not, calculation.spellName) then
			return
		end
		if index then
			if data.SelfCast then
				local unit = select(8,UnitDebuff("target",index))
				if not unit or not UnitIsUnit("player",unit) then return end
			end
			if data.SelfCastBuff then
				local unit = select(8,UnitBuff((target or "target"),index))
				if not unit or not UnitIsUnit("player",unit) then return end
			end
		end
		--Process active aura table
		if data.ActiveAura then
			if data.Ranks then
				ActiveAuras[data.ActiveAura] = rank and tonumber(string_match(rank,"%d+")) or data.Ranks
			elseif data.Index then
				ActiveAuras[data.ActiveAura] = index or 0
			else
				if apps and apps > 0 then
					ActiveAuras[data.ActiveAura] = apps
				else
					ActiveAuras[data.ActiveAura] = data.Apps or (ActiveAuras[data.ActiveAura] or 0) + 1
				end
			end
		end
		--Process category
		if data.Category and not data.SkipCategory then
			if ActiveAuras[data.Category] then return
			else ActiveAuras[data.Category] = true end
		end
		--Halt if incorrect texture
		if data.Texture and texture and not string_find( texture, data.Texture ) then return end
		local modType = data.ModType
		if modType == "Special" and DrDamage.Calculation[buffName] then
			apps = apps or data.Apps
			rank = rank and tonumber(string_match(rank,"%d+")) or data.Ranks
			DrDamage.Calculation[buffName]( calculation, ActiveAuras, Talents, index, apps, texture, rank )
		else			
			--NOTE: For manually added buffs
			if not index and data.Mods then
				rank = rank and tonumber(string_match(rank,"%d+")) or data.Ranks
				if not UnitBuff("player", buffName) and (not data.Alt or data.Alt and not UnitBuff("player", data.Alt)) then
					for k, v in pairs(data.Mods) do
						if calculation[k] then
							if type(v) == "function" then
								calculation[k] = v(calculation[k],rank)
							else
								calculation[k] = calculation[k] + v
							end
						end
					end
				end				
			end
			local value = data.Value
			if not value then return end	
			if data.Ranks then
				rank = rank and tonumber(string_match(rank,"%d+")) or data.Ranks
				value = (type(value) == "table") and value[rank] or value * rank
			end
			if data.Apps then
				apps = apps or data.Apps
				value = (type(value) == "table") and value[apps] or value * apps
			end		
			if not modType then
				calculation.dmgM = calculation.dmgM * (1 + value)			
			elseif calculation[modType] then
				if data.Multiply then
					calculation[modType] = calculation[modType] * (1 + value)
				else	
					calculation[modType] = calculation[modType] + value
				end
			end			
		end
	end
end

--Credits to the author of RatingBuster (Whitetooth) for the formula!
local ratingTypes = { ["Hit"] = 8, ["Crit"] = 14, ["MeleeHit"] = 10, ["SpellHaste"] = 10, ["Resilience"] = 28.75 }
function DrDamage:GetRating( rType, convertR, full )
	local playerLevel = UnitLevel("player")
	local base = ratingTypes[rType]
	local rating, value

	if playerLevel > 70 then
		rating = base * (82/52) * ((131/63)^((playerLevel - 70)/10))
	elseif playerLevel > 60 then
		rating = base * 82 / (262 - 3 * playerLevel)
	elseif playerLevel > 10 then
		rating = base * ((playerLevel - 8) / 52)
	elseif 	playerLevel <= 10 then
		rating = base / 26
	end

	value = convertR and convertR/rating or rating
	value = full and value or DrD_Round(value,2)
	return value
end

function DrDamage:GetLevels()
	local targetLevel = UnitLevel("target")
	local playerLevel = UnitLevel("player")
	local boss
	
	if (UnitClassification("target") == "worldboss") then
		targetLevel = playerLevel + 3
		boss = true
	elseif targetLevel == -1 then
		targetLevel = playerLevel + 10
	end
	
	return math_min(10, targetLevel - playerLevel), playerLevel, targetLevel, boss
end

--CORE: Base spell hit chance
local hitDataMOB = { [-4] = 100, [-3] = 99, [-2] = 98, [-1] = 97, [0] = 96, 95, 94, 83, 72, 61, 50, 39, 28, 17, 6 }
local hitDataPlayer = { [-4] = 100, [-3] = 99, [-2] = 98, [-1] = 97, [0] = 96, 95, 94, 87, 80, 73, 66, 59, 52, 45, 38 }
local hitMod = (select(2, UnitRace("player")) == "Draenei") and 1 or 0
function DrDamage:GetSpellHit()
	local delta = math_max(-4,settings.HitTarget and UnitExists("target") and (self:GetLevels()) or settings.TargetLevel)
	local npc = settings.HitTarget and not UnitIsPlayer("target") or not settings.HitTarget
	return (npc and hitDataMOB[delta] or hitDataPlayer[delta]) + hitMod + GetCombatRatingBonus(8)
end

local WeaponBuffScan = GetTime()
local WeaponBuff, WeaponBuffRank 
function DrDamage:GetWeaponBuff(off)
	local mh, _, _, oh = GetWeaponEnchantInfo()
	local name, rank, buff
	
	if not off and mh then
		if GetTime() > WeaponBuffScan then
			WeaponBuffScan = GetTime() + 2
			GT:SetInventoryItem("player", 16)
			_, _, buff = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
			if buff then
				name, rank = string_match(buff,"^(.*) (%d+)$")
			end
			WeaponBuff, WeaponBuffRank = name or buff, rank
		end
		return WeaponBuff, WeaponBuffRank
	elseif off and oh then
		GT:SetInventoryItem("player", 17)
		_, _, name, rank = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
		if buff then
			name, rank = string_match(buff,"^(.*) (%d+)$")
		end
	end
	return name or buff, rank
end

function DrDamage:Calc(name, rank, tooltip, modify, debug)
	if not spellInfo or not name then return end
	if not spellInfo[name] then return end
	local baseSpell = settings.SwapCalc and spellInfo[name]["Secondary"] and spellInfo[name]["Secondary"][0] or spellInfo[name][0]
	if type(baseSpell) == "function" then baseSpell = baseSpell() end
	if baseSpell then
		if playerCaster or playerHybrid and not baseSpell.Melee then
			return self:CasterCalc(name, rank, tooltip, modify, debug)								
		elseif playerMelee or playerHybrid and baseSpell.Melee then
			return self:MeleeCalc(name, rank, tooltip, modify, debug)
		end
	end
end

function DrDamage:TestAll()
	debug_flag = true
	local spells, ranks = 0, 0
	self:UpdateTalents(false,true)
	for k, v in pairs( spellInfo ) do
		local baseSpell = spellInfo[k][0]
		for i=1,#spellInfo[k] do
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterCalc(k, i, true, nil, true)
				if not baseSpell.NoTooltip then self:CasterTooltip( GameTooltip, k, i ) end
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeCalc(k, i, true, nil, true)
				if not baseSpell.NoTooltip then self:MeleeTooltip( GameTooltip, k, i ) end
			end
			ranks = ranks + 1
		end
		spells = spells + 1
	end
	debug_flag = false
	self:Print( "Tested: " .. ranks .. " ranks of " .. spells .. " spells.")
end

function DrDamage:Debug()
	debug_flag = not debug_flag
	self:Print( "Debug: ", debug_flag )
end

local calctable = {}
local auratable = {}
local talenttable = {}
function DrDamage:GetModifier(school, Talents)
	DrD_ClearTable( auratable )
	calctable.school = school
	--calctable.group = group
	--calctable.subType = subType	
	calctable.dmgM = 1
	calctable.dmgM_dot = 1
	calctable.dmgM_Extra = 1
	calctable.critPerc = 0
	calctable.critM = 0
	
	if next(settings["PlayerAura"]) then
		for buffName in pairs(settings["PlayerAura"]) do
			local aura = self.PlayerAura[buffName]
			if aura and not aura.Melee then
				self.BuffCalc( aura, calctable, auratable, Talents or talenttable, nil, buffName )
			end
		end
	end
	if next(settings["TargetAura"]) then
		for buffName in pairs(settings["TargetAura"]) do
			local aura = self.TargetAura[buffName]
			if aura and not aura.Melee then
				self.BuffCalc( aura, calctable, auratable, Talents or talenttable, nil, buffName )
			end
		end
	end	
	for index=1,40 do
		local buffName, rank, texture, apps = UnitBuff("player",index)
		if buffName then
			local aura = self.PlayerAura[buffName]
			if aura and not settings["PlayerAura"][buffName] and not aura.Update and not aura.Melee then
				self.BuffCalc( aura, calctable, auratable, Talents or talenttable, nil, buffName, nil, apps, texture, rank )
			end
		else break end
	end
	for index=1,40 do
		local buffName, rank, texture, apps = UnitDebuff("target",index)
		if buffName then
			local aura = self.TargetAura[buffName]
			if aura and not settings["TargetAura"][buffName] and not aura.Update and not aura.Melee then
				self.BuffCalc( aura, calctable, auratable, Talents or talenttable, nil, buffName, nil, apps, texture, rank )
			end
		else break end
	end
	return calctable.dmgM, calctable.critPerc
end

--[[
--Used for harvesting base value increases:
function DrDamage:HarvestValues(id, rk)
	local name, rank
	if rk then
		GT:SetHyperlink(GetSpellLink(id, rk))
		name, rank = GetSpellInfo(id, rk)
	else
		GT:SetHyperlink(GetSpellLink(id))
		name, rank = GetSpellInfo(id)
	end
	rank = rank and tonumber(string_match(rank,"%d+")) --or 1
	local min, max = string_match(GT:GetLine(GT:NumLines()),"(%d+) to (%d+)")
	if not min then 
		min = string_match(GT:GetLine(GT:NumLines()),"causing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"Transfers (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"Causes (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"doing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"absorbing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"Absorbs (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"deals (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"that causes (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"and causes (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"for (%d+) every")
		min = min or string_match(GT:GetLine(GT:NumLines()),"for (%d+) ")
		--min = min or string_match(GT:GetLine(GT:NumLines()),"(%d+)%.")
	end
	if not max then
		max = min
	end
	self:Print(name, rank, min, max)
	return name, rank, min, max
end

function DrDamage:Wowhead()
	local stuff = ""
	local table = {}
	local i = 1
	for k in string.gmatch(stuff,"spell=(%d+)") do
		table[i] = k
		i = i + 1
	end
	return table
end

local harvesttable = {}
local ranktable = {}
function DrDamage:Harvest(input,remove,rpercent,secondary,force)
	if input and input == 1 then
		input = DrDamage:Wowhead()
	end
	if input then
		local error
		if not remove then remove = 0 end
		if not rpercent then rpercent = 0 end
		DrD_ClearTable( harvesttable )
		DrD_ClearTable( ranktable )
		local entries = 0
		if (type(input) == "table") then
			self:Print("Using wowhead IDs")
			ranktable = input
		elseif GetSpellInfo(input,"Rank 1") then
			self:Print("Using spellbook ranks")
			for i = 1, 20 do
				if GetSpellInfo(input, "Rank " .. i) then
					ranktable[i] = "Rank " .. i
				end
			end
		else
			self:Print("Using spellID iterate")
			local j = 1
			for i=1,100000 do
				local sname, srank = GetSpellInfo(i)
				srank = srank and string_match(srank,"%d+")
				if input == sname and srank then
					ranktable[j] = i
					j = j + 1
				end
			end
		end
		for _, k in ipairs(ranktable) do
			entries = entries + 1
			local name, rank, min, max
			if (type(k) == "string") then
				name, rank, min, max = DrDamage:HarvestValues(input, k)
				ChatFrame1:AddMessage(GetSpellLink(input, k))
			else
				name, rank, min, max = DrDamage:HarvestValues(k)
				ChatFrame1:AddMessage(GetSpellLink(k))
			end
			local data = secondary and DrDamage.spellInfo[name] and DrDamage.spellInfo[name]["Secondary"] or DrDamage.spellInfo[name]
			if name and rank and data and data[rank] and min and max then
				local oldmin, oldmax, spellLevel = data[rank][1], data[rank][2]
				if harvesttable[rank] and ((harvesttable[rank][1] ~= min) or (harvesttable[rank][2] ~= max)) then
					self:Print("WARNING: Rank " .. rank .. " already exists in table with different values.")
					error = true
				else
					harvesttable[rank] = { min, max, oldmin, oldmax, data[rank] }
				end
			end
		end
		self:Print("NOTE: Spell ID entries: " .. entries)
		if #harvesttable > 0 then
			for i=1,#harvesttable do
				local val = harvesttable[i]
				if val then
					self:Print( "Rank " .. i .. ": " .. val[1] .. "-" .. val[2] .. " from " .. val[3] .. "-" .. val[4] )
				else
					self:Print( "Rank " .. i .. ": Error")
					error = true
				end
					
			end
			if not error or force then
				self:Print("Difference table:")
				for i=1,#harvesttable do
					local val = harvesttable[i]
					local rankstring = "[" .. i .. "]" .. " = { " .. val[3] .. ", " .. val[4] .. ", " .. (val[1] - val[3] - remove - rpercent * val[3]) .. ", " .. (val[2] - val[4] - remove - rpercent * val[4]) 
					for k, v in pairs(val[5]) do
						if type(k) ~= "number" then
							rankstring = rankstring .. ", " .. k .. " = " .. v
						end
					end
					rankstring = rankstring .. ", },"
					ChatFrame1:AddMessage( rankstring ) 
				end
			end
		end
	end
end
--]]